"""
 Copyright 2023. Aubin Ramon and Pietro Sormanni. CC BY-NC-SA 4.0
"""


import numpy

chemical_liabilities = {
# HD : sequence motif,  specific location only, location type  liability score Int , fine liability score, reference
'NxS' : [ 'N-glycosylation', None,'any', 1 , 0.5 , 'Jarasch, A. et al. Developability Assessment During the Selection of Novel Therapeutic Antibodies. J. Pharm. Sci. 104, (2015).' ],
'NxT' : [ 'N-glycosylation', None,'any', 1 , 0.5 , 'Jarasch, A. et al. Developability Assessment During the Selection of Novel Therapeutic Antibodies. J. Pharm. Sci. 104, (2015).' ],
#Prone to deamidation, asparagine and aspartic acid in more flexible regions, such as the CDR loops, are more prone, respectively, to deamidation and isomerisation than those found in rigid regions, such as the framework
'SNG' : [ 'deamidation',     None,'any', 2 ,  1 , 'Chelius D, Rehder DS, Bondarenko PV Identification and characterization of deamidation sites in the conserved regions of human immunoglobulin antibodies. Anal Chem. 2005'],
'ENN' : [ 'deamidation',     None,'any', 1 ,  0.6 , 'Chelius D, Rehder DS, Bondarenko PV Identification and characterization of deamidation sites in the conserved regions of human immunoglobulin antibodies. Anal Chem. 2005'],
'LNG' : [ 'deamidation',     None,'any', 2 ,  1 , 'Chelius D, Rehder DS, Bondarenko PV Identification and characterization of deamidation sites in the conserved regions of human immunoglobulin antibodies. Anal Chem. 2005'],
'LNN' : [ 'deamidation',     None,'any', 1 ,  0.6 , 'Chelius D, Rehder DS, Bondarenko PV Identification and characterization of deamidation sites in the conserved regions of human immunoglobulin antibodies. Anal Chem. 2005'],
# Isomerization, DG is expected to be more labile than DS, and DS more so than DD or DT
'DG' : [ 'isomerization',     None,'any', 1 ,  1 , 'Lu et al., Deamidation and isomerization liability analysis of 131 clinical-stage antibodies 2019'],
'DS' : [ 'isomerization',     None,'any', 1 ,  0.7 , 'Lu et al., Deamidation and isomerization liability analysis of 131 clinical-stage antibodies 2019'],
'DT' : [ 'isomerization',     None,'any', 0.5 ,  0.5, 'Lu et al., Deamidation and isomerization liability analysis of 131 clinical-stage antibodies 2019'],
'DD' : [ 'isomerization',     None,'any', 0.5 ,  0.5 , 'Lu et al., Deamidation and isomerization liability analysis of 131 clinical-stage antibodies 2019'],
'DH' : [ 'isomerization',     None,'any', 0.5 ,  0.5 , 'Lu et al., Deamidation and isomerization liability analysis of 131 clinical-stage antibodies 2019'],
# Asparagine residues followed by certain sequence motifs including glycine, serine, threonine, aspartate and histidine have been reported as degradation hotspots
'NG' : [ 'deamidation',     None,'any', 1 ,  0.7 , 'Lu et al., Deamidation and isomerization liability analysis of 131 clinical-stage antibodies 2019'],
'NS' : [ 'deamidation',     None,'any', 0.5 ,  0.3 , 'Lu et al., Deamidation and isomerization liability analysis of 131 clinical-stage antibodies 2019'],
'NT' : [ 'deamidation',     None,'any', 0.5 ,  0.2 , 'Lu et al., Deamidation and isomerization liability analysis of 131 clinical-stage antibodies 2019'],
'NH' : [ 'deamidation',     None,'any', 0.5 ,  0.2 , 'Lu et al., Deamidation and isomerization liability analysis of 131 clinical-stage antibodies 2019'],
'ND' : [ 'deamidation',     None,'any', 0.5 ,  0.2 , 'Lu et al., Deamidation and isomerization liability analysis of 131 clinical-stage antibodies 2019']
}

cdr_liabilities = {
# HD : liability, sequence motif,  specific location only,  liability score Int , fine liability score incremental, reference
# Cross-interaction: RS highest, WS less, YG even less
'RS' : [ 'poor-specificity',     'CDR','CDR', 1 ,  0.9, 'Kelly et al., 2018 Reduction of Nonspecificity Motifs in Synthetic Antibody Libraries'],
'WS' : [ 'poor-specificity',     'CDR','CDR', 1 ,  0.8, 'Kelly et al., 2018 Reduction of Nonspecificity Motifs in Synthetic Antibody Libraries'],
'YG' : [ 'poor-specificity',     'CDR','CDR', 0.5 ,  0.5, 'Kelly et al., 2018 Reduction of Nonspecificity Motifs in Synthetic Antibody Libraries'],
# Enrichment in CDR H3 causes sometimes but rarely non-specificty. Glycine contributes to flexibility of H3 loop, which may allow greater olasticity and broader antigen recognition
'n3G' : [ 'poor-specificity',   'CDRH3','CDR', 0.5 ,  0.3, 'Kelly et al., 2018 Reduction of Nonspecificity Motifs in Synthetic Antibody Libraries'], # n2 means 2 or more
'n2V' : [ 'poor-specificity',   'CDRH3','CDR', 0.5 ,  0.2, 'Kelly et al., 2018 Reduction of Nonspecificity Motifs in Synthetic Antibody Libraries'],
'n1W' : [ 'poor-specificity',   'CDRH3','CDR', 1 ,  0.6, 'Kelly et al., 2018 Reduction of Nonspecificity Motifs in Synthetic Antibody Libraries'],
'n1C' : [ 'disulphide-dimerisation','CDR','CDR',1,  1  , 'Kelly et al., 2018 Reduction of Nonspecificity Motifs in Synthetic Antibody Libraries'],
'n1M' : [ 'oxydation','CDR','CDR',1,  1  , ''],
'n2R' : [ 'poor-specificity',   'CDRH3','CDR', 0.5 ,  0.3, 'Kelly et al., 2018 Reduction of Nonspecificity Motifs in Synthetic Antibody Libraries'],
# The best indicators of antibody specificity in terms of CDR amino acid composition are reduced levels of arginine and lysine. Lysines are very rare in CDRs, but could potentially be a glycation site and reduced activity
'n2K' : [ 'poor-specificity/glycation','CDR','CDR',0.5,1  , 'Kelly et al., 2018 Reduction of Nonspecificity Motifs in Synthetic Antibody Libraries']
#K49    Glycation hotspot    CDR2LC    https://pubmed.ncbi.nlm.nih.gov/18307322/
#K98    Glycation hotspot    CDR3HC    https://pubmed.ncbi.nlm.nih.gov/21287557/
#M4    Critical quality attributes.  Kabat numbering. Employed  specific stress conditions to antibody, elevated temperatures, pH, oxidizing agents, and forced glycation with glucose incubation. However, none of the assessed degradation products led to a complete loss of functionality if only one light or heavy chain of the native antibody was affected    LC    https://pubmed.ncbi.nlm.nih.gov/24441081/
#N30/31    Critical quality attributes.  Kabat numbering. Employed  specific stress conditions to antibody, elevated temperatures, pH, oxidizing agents, and forced glycation with glucose incubation. However, none of the assessed degradation products led to a complete loss of functionality if only one light or heavy chain of the native antibody was affected    LC    https://pubmed.ncbi.nlm.nih.gov/24441081/
#N92    Critical quality attributes.  Kabat numbering. Employed  specific stress conditions to antibody, elevated temperatures, pH, oxidizing agents, and forced glycation with glucose incubation. However, none of the assessed degradation products led to a complete loss of functionality if only one light or heavy chain of the native antibody was affected    LC    https://pubmed.ncbi.nlm.nih.gov/24441081/
#M100c    Critical quality attributes.  Kabat numbering. Employed  specific stress conditions to antibody, elevated temperatures, pH, oxidizing agents, and forced glycation with glucose incubation. However, none of the assessed degradation products led to a complete loss of functionality if only one light or heavy chain of the native antibody was affected    HC     https://pubmed.ncbi.nlm.nih.gov/24441081/
#K33    Critical quality attributes. Kabat numbering. Employed  specific stress conditions to antibody, elevated temperatures, pH, oxidizing agents, and forced glycation with glucose incubation. However, none of the assessed degradation products led to a complete loss of functionality if only one light or heavy chain of the native antibody was affected    HC    https://pubmed.ncbi.nlm.nih.gov/24441081/
#H54    Residues account for most chemical modifications in 121 clinical antibody sequences, Kabat-Chotia numbering
#H98    Residues account for most chemical modifications in 121 clinical antibody sequences, Kabat-Chotia numbering
#L30    Residues account for most chemical modifications in 121 clinical antibody sequences, Kabat-Chotia numbering
}



import math
import os
from collections import defaultdict
from typing import Tuple

import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
import pandas as pd
import seaborn as sns
import torch
import pytorch_lightning as pl
from sklearn.metrics import roc_curve, auc, precision_recall_curve
import textwrap

from .alignment.aho_consensus import (
    cdr1_aho_indices,
    cdr2_aho_indices,
    cdr3_aho_indices,
    fr_aho_indices,
)
from .alignment.mybio import anarci_alignments_of_Fv_sequences_iter
from .onehotencoder import (
    alphabet,
    data_loader_masking_bert_onehot_csv_paired,
    data_loader_masking_bert_onehot_fasta,
    data_loader_masking_bert_onehot_csv_paired_with_mismatch,
)


labelled_colors = {
    "pssm-generated": sns.color_palette("tab10")[1],
    "validation": sns.color_palette("tab10")[3],
    "mouse": sns.color_palette("tab10")[4],
    "human": sns.color_palette("tab10")[5],
    "rhesus": sns.color_palette("tab10")[6],
}

fontsizes = {
    "x_label": 16,
    "y_label": 16,
    "title": 20,
    "legend": 16,
}


def process_ROC(dataframes, labels, output_filepath, title, model_type=None):
    # Load the original dataset
    original_df = dataframes[0]
    original_scores = original_df[f"AbNatiV {model_type} Score"]

    # List of new datasets to compare
    new_dataframes = dataframes[1:]
    new_labels = labels[1:]

    # Initialize a plot
    plt.figure(figsize=(7, 6))

    roc_aucs = []

    for i, new_df in enumerate(new_dataframes):
        # Load the new dataset
        new_scores = new_df[f"AbNatiV {model_type} Score"]

        # Combine scores and create true labels
        combined_scores = pd.concat([original_scores, new_scores])

        true_labels = pd.Series([1] * len(original_scores) + [0] * len(new_scores))

        # Calculate ROC curve and AUC
        fpr, tpr, _ = roc_curve(true_labels, combined_scores)
        roc_auc = auc(fpr, tpr)
        roc_aucs.append(roc_auc)

        label = new_labels[i]
        color = labelled_colors[label.lower()]

        # Plot ROC curve
        plt.plot(fpr, tpr, label=f"{label}: {roc_auc:.3f} AUC", color=color, lw=3)

    # Plot the diagonal line
    plt.plot([0, 1], [0, 1], color="navy", lw=2, linestyle="--")

    # Set plot limits and labels
    plt.xlim([0.0, 1.05])
    plt.ylim([0.0, 1.05])
    # Set plot limits and labels
    sns.despine()

    plt.xlabel("False Positive Rate", fontsize=16)
    plt.ylabel("True Positive Rate", fontsize=16)
    plt.title(title)

    legend = plt.legend(loc="lower right", fontsize=16, frameon=False)

    if output_filepath:
        output_dir = os.path.dirname(output_filepath)
        if not os.path.exists(output_dir):
            os.makedirs(output_dir, exist_ok=True)

        # Save the plot as a PNG file
        plt.savefig(output_filepath, dpi=400, bbox_inches="tight")
    else:
        plt.show()

    plt.close()

    return roc_aucs


def process_PR(dataframes, labels, output_filepath=None, title="", model_type="None"):
    # Load the original dataset
    original_df = dataframes[0]
    original_scores = original_df[f"AbNatiV {model_type} Score"]

    # List of new datasets to compare
    new_dataframes = dataframes[1:]
    new_labels = labels[1:]

    # Initialize a plot
    plt.figure(figsize=(6, 6))
    plt.rcParams["font.family"] = "sans-serif"

    baselines = {}

    pr_aucs = []

    for i, new_df in enumerate(new_dataframes):
        new_scores = new_df[f"AbNatiV {model_type} Score"]

        # Combine scores and create true labels
        combined_scores = pd.concat([original_scores, new_scores])
        true_labels = pd.Series([1] * len(original_scores) + [0] * len(new_scores))

        # Calculate AUC
        precision, recall, thresholds = precision_recall_curve(
            true_labels, combined_scores
        )
        pr_auc = auc(recall, precision)
        pr_aucs.append(pr_auc)

        f1_scores = 2 * (precision * recall) / (precision + recall)

        # Find the threshold that gives the maximum F1-score
        best_index = f1_scores.argmax()
        best_threshold = thresholds[best_index]
        best_f1 = f1_scores[best_index]

        label = new_labels[i]
        color = labelled_colors[label.lower()]
        # Plot ROC curve
        plt.plot(
            recall, precision, label=f"{label}: {pr_auc:.3f} AUC", color=color, lw=3
        )

        plt.scatter(
            recall[best_index],
            precision[best_index],
            color="black",
            label=f"Best Threshold {best_threshold:.4f} (F1 = {best_f1:.2f})",
        )
        print(label)
        print(f"Best Threshold {best_threshold}\n(F1 = {best_f1:.2f})\n")
        baseline = round(len(true_labels[true_labels == 1]) / len(true_labels), 3)
        if baseline not in baselines:
            baselines[baseline] = []
        baselines[baseline].append(label)

    for baseline, labels in baselines.items():
        if len(baselines) > 1:
            label = f'Baseline {" - ".join(labels)}: {baseline:.3f} AUC'
        else:
            label = f"Baseline: {baseline:.3f} AUC"
        color = (
            labelled_colors[labels[0].lower()]
            if len(baselines) > 1
            else (128 / 255, 128 / 255, 128 / 255, 0.5)
        )
        plt.plot(
            [0, 1], [baseline, baseline], linestyle="--", color=color, label=label, lw=2
        )

    # Set plot limits and labels
    plt.xlim([-0.025, 1.025])
    plt.ylim([-0.025, 1.025])

    # Set plot limits and labels
    sns.despine(right=True, top=True)

    plt.xlabel("Recall", fontsize=fontsizes["x_label"])
    plt.ylabel("Precision", fontsize=fontsizes["y_label"])
    wrapped_title = "\n".join(textwrap.wrap(title, width=20))
    plt.suptitle(wrapped_title, y=0.45, fontsize=fontsizes["title"])

    legend = plt.legend(loc="lower left", frameon=False, fontsize=fontsizes["legend"])

    if output_filepath:
        output_dir = os.path.dirname(output_filepath)
        if not os.path.exists(output_dir):
            os.makedirs(output_dir, exist_ok=True)

        # Save the plot as a PNG file
        plt.savefig(output_filepath, dpi=400, bbox_inches="tight")
    else:
        plt.show()

    plt.close()

    return pr_aucs


# Copied from scoring_functions to avoid circular imports!
# TODO: tidy and refactor
def norm_by_length_no_gap(scores_pposi: list, onehot_encoding: list) -> list:
    """
    Sum the scores per position of a given sequence. Normalise the result by the
    number of residues (no gaps).

    Parameters
    ----------
        - scores_pposi : list
            Scores of the chosen positions
        - onehot_encoding : list
            One-hot encoding of the chosen given positions
    """
    length = len(onehot_encoding)
    idx = np.argmax(onehot_encoding, axis=1)
    nb_gaps = np.count_nonzero(idx == 20)
    length -= nb_gaps
    if length == 0:
        print("Length portion equals zero, too many gaps. Evaluates it as np.nan")
        norm = np.nan
    else:
        norm = np.sum(scores_pposi) / length
    return norm


def linear_rescaling(list_scores: list, t_new: int, t_r: int) -> list:
    """
    Linear rescaling of the scores to translate the threshold t_r (specific to each
    model type as defined in Methods) to the new threshold t_new.
    """
    rescaled = list()
    for x in list_scores:
        rescaled.append((t_new - 1) / (t_r - 1) * (x - 1) + 1)
    return rescaled


def get_abnativ_nativeness_scores(
    output_abnativ: dict, portion_indices: list, model_type: str
) -> list:
    """
    Give the AbNatiV nativeness scores for given positions (could be full) of the outputs of the
    AbNatiV model.

    Parameters
    ----------
        - output_abnativ : dict
            Output dict of sequences evaluated by AbNatiV
            i.e. {'fp':'/my/path/to/the/dataset.txt', 'recon_error_pbe': [0.2,0.3,0.4]}
        - portion_indices : list
            Position indices to score
            e.g., range(1,150) for the all sequence / range(27,43) for the CDR-1 (AHo numbering)
        - model_type : str
            VH, VHH, VKappa, VLambda

    Returns
    ----------
        - a list of the rescaled AbNatiV nativenees scores
    """

    humanness_scores = list()
    best_thresholds = {}

    # Convert AHo-position into list index
    portion_indices = np.array(portion_indices) - 1

    # Score
    for k, profile in enumerate(output_abnativ["mse_pposi"]):
        score = norm_by_length_no_gap(
            profile[portion_indices].cpu().detach().numpy(),
            output_abnativ["inputs"][k][portion_indices].cpu().detach().numpy(),
        )
        humanness_scores.append(math.exp(-score))

    if model_type not in best_thresholds.keys():  # If scoring your own model
        rescaled_scores = humanness_scores

    else:
        rescaled_scores = linear_rescaling(
            humanness_scores, 0.8, best_thresholds[model_type]
        )

    return list(rescaled_scores)


def get_abnativ_scores_per_batch(
    model: pl.LightningModule,
    model_type: str,
    batch: int,
    scored_data_dict_mean: dict,
    aho_posi_whole=range(1, 150),
) -> dict:
    """
    Evaluate a batch of sequences using an AbNatiV model and update the score dictionary with nativeness scores.

    Parameters
    ----------
        - model : pl.LightningModule
            An instance of the AbNatiV_Model class used to evaluate the sequences.
        - model_type : str
            VH, VHH, VKappa, VLambda
        - batch : list
            A list of sequences to be evaluated.
        - scored_data_dict_mean : dict
            A dictionary to be updated with the new batch data. It should be a defaultdict(list) or similar structure.

    Returns
    -------
        - A dictionary containing the output of the sequences evaluated by AbNatiV.
            Example: {'fp': '/my/path/to/the/dataset.txt', 'recon_error_pbe': [0.2, 0.3, 0.4]}
    """

    output_abnativ = model(batch)
    name_type = model_type

    # Sequence-level scores
    humanness_scores = get_abnativ_nativeness_scores(
        output_abnativ, aho_posi_whole, model_type
    )
    cdr1_scores = get_abnativ_nativeness_scores(
        output_abnativ, cdr1_aho_indices, model_type
    )
    cdr2_scores = get_abnativ_nativeness_scores(
        output_abnativ, cdr2_aho_indices, model_type
    )
    cdr3_scores = get_abnativ_nativeness_scores(
        output_abnativ, cdr3_aho_indices, model_type
    )
    fr_scores = get_abnativ_nativeness_scores(
        output_abnativ, fr_aho_indices, model_type
    )

    scored_data_dict_mean[f"AbNatiV {name_type} Score"].extend(humanness_scores)

    scored_data_dict_mean[f"AbNatiV CDR1-{name_type} Score"].extend(cdr1_scores)
    scored_data_dict_mean[f"AbNatiV CDR2-{name_type} Score"].extend(cdr2_scores)
    scored_data_dict_mean[f"AbNatiV CDR3-{name_type} Score"].extend(cdr3_scores)
    scored_data_dict_mean[f"AbNatiV FR-{name_type} Score"].extend(fr_scores)

    return output_abnativ


def evaluate_pssm(
    model: pl.LightningModule, model_type: str, fp_pssm: str, batch_size: int
):
    """
    Evaluate PSSM scores from the specified file path using the provided model and return the results as a pandas DataFrame.

    Parameters
    ----------
        - model : pl.LightningModule
            The model used to evaluate the PSSM scores.
        - model_type : str
            VH, VHH, VKappa, VLambda
        - fp_pssm : str
            File path to the PSSM data in FASTA format.
        - batch_size : int
            Batch size for loading the PSSM data.

    Returns
    -------
        - pd.DataFrame
            A DataFrame containing the evaluated PSSM scores.
    """
    scored_data_dict_mean = defaultdict(list)

    pssm_loader = data_loader_masking_bert_onehot_fasta(
        fp_pssm, batch_size, perc_masked_residues=0, is_masking=False
    )

    # TODO: Consider better device handling
    device = next(model.parameters()).device

    for batch in pssm_loader:
        batch = [tensor.to(device) for tensor in batch]
        output_abnativ = get_abnativ_scores_per_batch(
            model=model,
            model_type=model_type,
            batch=batch,
            scored_data_dict_mean=scored_data_dict_mean,
        )

    df_mean = pd.DataFrame.from_dict(scored_data_dict_mean)

    return df_mean


def move_to_device(data, device):
    if isinstance(data, torch.Tensor):
        return data.to(device)
    elif isinstance(data, list):
        return [move_to_device(item, device) for item in data]
    elif isinstance(data, tuple):
        return tuple(move_to_device(item, device) for item in data)
    else:
        raise TypeError(f"Unsupported data type: {type(data)}")


def evaluate_paired_pssm(
    model: pl.LightningModule,
    model_type: str,
    fp_pssm: str,
    batch_size: int,
    aho_posi_whole=range(1, 299),
):
    """
    Evaluate PSSM scores from the specified file path using the provided model and return the results as a pandas DataFrame.
    For paired sequences when loading the inputs. (using a paired csv file)
    Parameters
    ----------
        - model : pl.LightningModule
            The model used to evaluate the PSSM scores.
        - model_type : str
            VH, VHH, VKappa, VLambda
        - fp_pssm : str
            File path to the PSSM data in FASTA format.
        - batch_size : int
            Batch size for loading the PSSM data.

    Returns
    -------
        - pd.DataFrame
            A DataFrame containing the evaluated PSSM scores.
    """
    scored_data_dict_mean = defaultdict(list)

    pssm_loader = data_loader_masking_bert_onehot_csv_paired(
        fp_pssm, batch_size=batch_size, perc_masked_residues=0, is_masking=False
    )

    # TODO: Consider better device handling
    device = next(model.parameters()).device

    for batch in pssm_loader:
        batch = move_to_device(batch, device)
        output_abnativ = get_abnativ_scores_per_batch(
            model=model,
            model_type=model_type,
            batch=batch,
            scored_data_dict_mean=scored_data_dict_mean,
            aho_posi_whole=aho_posi_whole,
        )

    df_mean = pd.DataFrame.from_dict(scored_data_dict_mean)

    return df_mean


def evaluate_paired_pssm_with_mismatch(
    model: pl.LightningModule,
    model_type: str,
    fp_pssm: str,
    batch_size: int,
    aho_posi_whole=range(1, 299),
):
    """
    Evaluate PSSM scores from the specified file path using the provided model and return the results as a pandas DataFrame.
    For paired sequences when loading the inputs. (using a paired csv file)
    Parameters
    ----------
        - model : pl.LightningModule
            The model used to evaluate the PSSM scores.
        - model_type : str
            VH, VHH, VKappa, VLambda
        - fp_pssm : str
            File path to the PSSM data in FASTA format.
        - batch_size : int
            Batch size for loading the PSSM data.

    Returns
    -------
        - pd.DataFrame
            A DataFrame containing the evaluated PSSM scores.
    """
    scored_data_dict_mean = defaultdict(list)

    pssm_loader = data_loader_masking_bert_onehot_csv_paired_with_mismatch(
        fp_pssm,
        fp_pssm,
        batch_size=batch_size,
        perc_masked_residues=0,
        is_masking=False,
    )

    # TODO: Consider better device handling
    device = next(model.parameters()).device

    for batch in pssm_loader:
        batch = move_to_device(batch, device)
        output_abnativ = get_abnativ_scores_per_batch(
            model=model,
            model_type=model_type,
            batch=batch,
            scored_data_dict_mean=scored_data_dict_mean,
            aho_posi_whole=aho_posi_whole,
        )

    df_mean = pd.DataFrame.from_dict(scored_data_dict_mean)

    return df_mean


def evaluate_pairing_pred(
    model: pl.LightningModule,
    model_type: str,
    fp_pssm: str,
    batch_size: int,
    aho_posi_whole=range(1, 299),
):
    """
    Evaluate Pairing preiction for the top prediciton head
    Parameters
    ----------
        - model : pl.LightningModule
            The model used to evaluate the PSSM scores.
        - model_type : str
            VH, VHH, VKappa, VLambda
        - fp_pssm : str
            File path to the PSSM data in FASTA format.
        - batch_size : int
            Batch size for loading the PSSM data.

    Returns
    -------
        -  list
    """

    pssm_loader = data_loader_masking_bert_onehot_csv_paired(
        fp_pssm, batch_size=batch_size, perc_masked_residues=0, is_masking=False
    )

    # TODO: Consider better device handling
    device = next(model.parameters()).device

    saved_pairing_pred = list()
    for batch in pssm_loader:
        batch = move_to_device(batch, device)
        output_abnativ = model(batch)

        saved_pairing_pred += list(
            output_abnativ["pairing_pred"].cpu().detach().numpy()
        )

    return saved_pairing_pred


def evaluate_pairing_pred_with_mismatch(
    model: pl.LightningModule,
    model_type: str,
    fp_pssm: str,
    batch_size: int,
    aho_posi_whole=range(1, 299),
):
    """
    Evaluate Pairing preiction for the top prediciton head
    Parameters
    ----------
        - model : pl.LightningModule
            The model used to evaluate the PSSM scores.
        - model_type : str
            VH, VHH, VKappa, VLambda
        - fp_pssm : str
            File path to the PSSM data in FASTA format.
        - batch_size : int
            Batch size for loading the PSSM data.

    Returns
    -------
        -  list
    """

    pssm_loader = data_loader_masking_bert_onehot_csv_paired_with_mismatch(
        fp_pssm,
        fp_pssm,
        batch_size=batch_size,
        perc_masked_residues=0,
        is_masking=False,
    )

    # TODO: Consider better device handling
    device = next(model.parameters()).device

    saved_pairing_pred = list()
    for batch in pssm_loader:
        batch = move_to_device(batch, device)
        output_abnativ = model(batch)

        saved_pairing_pred += list(
            output_abnativ["pairing_pred"].cpu().detach().numpy()
        )

    return saved_pairing_pred


def evaluate_portion_score(
    model: pl.LightningModule,
    model_type: str,
    fp_database: str,
    batch_size: int,
    aho_selected_portion: list,
):
    """
    Evaluate portion of sequence of scores from the specified file path using the provided model and return the results as a pandas DataFrame.

    Parameters
    ----------
        - model : pl.LightningModule
            The model used to evaluate the PSSM scores.
        - model_type : str
            VH, VHH, VKappa, VLambda
        - fp_pssm : str
            File path to the PSSM data in FASTA format.
        - batch_size : int
            Batch size for loading the PSSM data.
        - aho_selected_portion : list of int
            List of AHo positions

    Returns
    -------
        - pd.DataFrame
            A DataFrame containing the evaluated PSSM scores.
    """
    scored_data_dict = defaultdict(list)

    pssm_loader = data_loader_masking_bert_onehot_fasta(
        fp_database, batch_size, perc_masked_residues=0, is_masking=False
    )

    # TODO: Consider better device handling
    device = next(model.parameters()).device

    for batch in pssm_loader:
        batch = [tensor.to(device) for tensor in batch]
        output_abnativ = model(batch)

        selected_scores = get_abnativ_nativeness_scores(
            output_abnativ, aho_selected_portion, model_type
        )

        scored_data_dict["AbNatiV selected portion score"].extend(selected_scores)

    df = pd.DataFrame.from_dict(scored_data_dict)

    return df


def calculate_pssm_aucs(
    model: pl.LightningModule,
    model_type: str,
    df_mean_val: pd.DataFrame,
    fp_pssm: str,
    batch_size: int,
    output_dir: str,
) -> Tuple[int, int]:
    """
    Evaluate PSSM scores, save the results, and calculate PR and ROC AUCs.

    This function performs the following steps:
    1. Randomly samples 10,000 examples from the validation scores if the dataset is larger.
    2. Evaluates PSSM scores from the provided file path.
    3. Saves the evaluated PSSM scores to 'pssm_scores.csv' in the specified directory.
    4. Processes and plots Precision-Recall (PR) curves for validation and PSSM-generated data.
    5. Processes and plots Receiver Operating Characteristic (ROC) curves for validation and PSSM-generated data.
    6. Returns the PR and ROC AUC scores.

    Parameters
    ----------
        - model : pl.LightningModule
            The model used to evaluate the PSSM scores.
        - model_type : str
            The type of model being used. Possible values are 'VH', 'VHH', 'VKappa', 'VLambda'.
        - df_mean_val : pd.DataFrame
            A DataFrame containing the mean validation scores.
        - fp_pssm : str
            File path to the PSSM data in FASTA format.
        - batch_size : int
            Batch size for loading the PSSM data.
        - output_dir : str
            The directory where the results will be saved.

    Returns
    -------
    tuple
        A tuple containing the PR AUC scores and ROC AUC scores.
    """

    # Random sample 10k examples from validation scores
    if len(df_mean_val) > 10000:
        df_mean_val = df_mean_val.sample(n=10000, random_state=11)

    # Evaluate PSSM scores from the provided file path and save the results in 'pssm_scores.csv' within the specified directory.
    df_mean_pssm = evaluate_pssm(model, model_type, fp_pssm, batch_size)
    df_mean_pssm.to_csv(f"{output_dir}/pssm_scores.csv")

    pr_aucs = process_PR(
        dataframes=[df_mean_val, df_mean_pssm],
        labels=["Validation", "PSSM-generated"],
        output_filepath=f"{output_dir}/PR_val_pssm.png",
        title="PR Validation-PSSM generated",
        model_type=model_type,
    )

    roc_aucs = process_ROC(
        dataframes=[df_mean_val, df_mean_pssm],
        labels=["Validation", "PSSM-generated"],
        output_filepath=f"{output_dir}/ROC_val_pssm.png",
        title="ROC Validation-PSSM generated",
        model_type=model_type,
    )

    return pr_aucs, roc_aucs


def calculate_pssm_aucs_paired(
    model: pl.LightningModule,
    model_type: str,
    df_mean_val: pd.DataFrame,
    fp_pssm: str,
    batch_size: int,
    output_dir: str,
) -> Tuple[int, int]:
    """
    Evaluate PSSM scores, save the results, and calculate PR and ROC AUCs. For paired sequences.

    This function performs the following steps:
    1. Randomly samples 10,000 examples from the validation scores if the dataset is larger.
    2. Evaluates PSSM scores from the provided file path.
    3. Saves the evaluated PSSM scores to 'pssm_scores.csv' in the specified directory.
    4. Processes and plots Precision-Recall (PR) curves for validation and PSSM-generated data.
    5. Processes and plots Receiver Operating Characteristic (ROC) curves for validation and PSSM-generated data.
    6. Returns the PR and ROC AUC scores.

    Parameters
    ----------
        - model : pl.LightningModule
            The model used to evaluate the PSSM scores.
        - model_type : str
            The type of model being used. Possible values are 'VH', 'VHH', 'VKappa', 'VLambda'.
        - df_mean_val : pd.DataFrame
            A DataFrame containing the mean validation scores.
        - fp_pssm : str
            File path to the PSSM data in FASTA format.
        - batch_size : int
            Batch size for loading the PSSM data.
        - output_dir : str
            The directory where the results will be saved.

    Returns
    -------
    tuple
        A tuple containing the PR AUC scores and ROC AUC scores.
    """

    # Random sample 10k examples from validation scores
    if len(df_mean_val) > 10000:
        df_mean_val = df_mean_val.sample(n=10000, random_state=11)
    if len(fp_pssm) > 10000:
        fp_pssm = fp_pssm.sample(n=10000, random_state=11)

    # Evaluate PSSM scores from the provided file path and save the results in 'pssm_scores.csv' within the specified directory.
    df_mean_pssm = evaluate_paired_pssm(model, model_type, fp_pssm, batch_size)
    df_mean_pssm.to_csv(f"{output_dir}/pssm_scores.csv")

    pr_aucs = process_PR(
        dataframes=[df_mean_val, df_mean_pssm],
        labels=["Validation", "PSSM-generated"],
        output_filepath=f"{output_dir}/PR_val_pssm.png",
        title="PR Validation-PSSM generated",
        model_type=model_type,
    )

    roc_aucs = process_ROC(
        dataframes=[df_mean_val, df_mean_pssm],
        labels=["Validation", "PSSM-generated"],
        output_filepath=f"{output_dir}/ROC_val_pssm.png",
        title="ROC Validation-PSSM generated",
        model_type=model_type,
    )

    return pr_aucs, roc_aucs


def calculate_pssm_aucs_paired_with_mismatch(
    model: pl.LightningModule,
    model_type: str,
    df_mean_val: pd.DataFrame,
    fp_pssm: str,
    batch_size: int,
    output_dir: str,
) -> Tuple[int, int]:
    """
    Evaluate PSSM scores, save the results, and calculate PR and ROC AUCs. For paired sequences.

    This function performs the following steps:
    1. Randomly samples 10,000 examples from the validation scores if the dataset is larger.
    2. Evaluates PSSM scores from the provided file path.
    3. Saves the evaluated PSSM scores to 'pssm_scores.csv' in the specified directory.
    4. Processes and plots Precision-Recall (PR) curves for validation and PSSM-generated data.
    5. Processes and plots Receiver Operating Characteristic (ROC) curves for validation and PSSM-generated data.
    6. Returns the PR and ROC AUC scores.

    Parameters
    ----------
        - model : pl.LightningModule
            The model used to evaluate the PSSM scores.
        - model_type : str
            The type of model being used. Possible values are 'VH', 'VHH', 'VKappa', 'VLambda'.
        - df_mean_val : pd.DataFrame
            A DataFrame containing the mean validation scores.
        - fp_pssm : str
            File path to the PSSM data in FASTA format.
        - batch_size : int
            Batch size for loading the PSSM data.
        - output_dir : str
            The directory where the results will be saved.

    Returns
    -------
    tuple
        A tuple containing the PR AUC scores and ROC AUC scores.
    """

    # Random sample 10k examples from validation scores
    if len(df_mean_val) > 10000:
        df_mean_val = df_mean_val.sample(n=10000, random_state=11)
    if len(fp_pssm) > 10000:
        fp_pssm = fp_pssm.sample(n=10000, random_state=11)

    # Evaluate PSSM scores from the provided file path and save the results in 'pssm_scores.csv' within the specified directory.
    df_mean_pssm = evaluate_paired_pssm_with_mismatch(
        model, model_type, fp_pssm, batch_size
    )
    df_mean_pssm.to_csv(f"{output_dir}/pssm_scores.csv")

    pr_aucs = process_PR(
        dataframes=[df_mean_val, df_mean_pssm],
        labels=["Validation", "PSSM-generated"],
        output_filepath=f"{output_dir}/PR_val_pssm.png",
        title="PR Validation-PSSM generated",
        model_type=model_type,
    )

    roc_aucs = process_ROC(
        dataframes=[df_mean_val, df_mean_pssm],
        labels=["Validation", "PSSM-generated"],
        output_filepath=f"{output_dir}/ROC_val_pssm.png",
        title="ROC Validation-PSSM generated",
        model_type=model_type,
    )

    return pr_aucs, roc_aucs



# (c) 2023 Sormannilab and Aubin Ramon
#
# To visualise the humanised variants with Chimera X
#
# ============================================================================
import os
from typing import List, Dict, Tuple
import Bio
from Bio.PDB import PDBParser


AA3_TO_1 = {
    "ALA":"A","ARG":"R","ASN":"N","ASP":"D","CYS":"C","GLN":"Q","GLU":"E","GLY":"G",
    "HIS":"H","ILE":"I","LEU":"L","LYS":"K","MET":"M","PHE":"F","PRO":"P","SER":"S",
    "THR":"T","TRP":"W","TYR":"Y","VAL":"V","SEC":"U","PYL":"O"
}

def _pdb_resname_to_1(resname: str) -> str:
    return AA3_TO_1.get(resname.upper().strip(), "X")

def _collect_chain_residues_aho(fp_pdb: str, chain_id: str) -> Dict[int, str]:
    """
    Read a (renumbered-to-AHo) PDB and return {resnum(AHo): one_letter_aa} for a given chain.
    Ignores insertion codes and non-standard residues.
    """
    parser = PDBParser(QUIET=True)
    structure = parser.get_structure("S", fp_pdb)
    model = structure[0]
    out = {}
    try:
        chain = model[chain_id]
    except KeyError:
        raise ValueError(f"Chain '{chain_id}' not found in {fp_pdb}")
    for res in chain.get_residues():
        het, resseq, icode = res.id  # e.g., (' ', 52, ' ')
        if het.strip():  # skip HETATM residues
            continue
        aa1 = _pdb_resname_to_1(res.get_resname())
        out[resseq] = aa1
    return out

def get_mutated_positions_from_aho_renumbered_pdbs(
    wt_pdb_aho: str, wt_chain: str,
    hum_pdb_aho: str, hum_chain: str
) -> List[Tuple[int, str, str]]:
    """
    Compare two AHo-renumbered PDBs and return list of (AHo_pos, wt_aa, hum_aa) where residues differ
    and both sides are standard amino acids.
    """
    wt_map = _collect_chain_residues_aho(wt_pdb_aho, wt_chain)
    hum_map = _collect_chain_residues_aho(hum_pdb_aho, hum_chain)
    mut_positions = []
    for pos in sorted(set(wt_map.keys()).intersection(hum_map.keys())):
        wt_aa, hum_aa = wt_map[pos], hum_map[pos]
        if wt_aa != "X" and hum_aa != "X" and wt_aa != hum_aa:
            mut_positions.append((pos, wt_aa, hum_aa))
    return mut_positions


def write_chimerax_cxc_visualisation(
    wt_pdb: str, wt_chain_h: str, wt_chain_l: str,
    hum_pdb: str, hum_chain_h: str, hum_chain_l: str,
    wt_pdb_crys: str, wt_chain_h_crys: str, wt_chain_l_crys: str,
    is_vhh: bool, 
    out_cxc_path: str,
    color_wt: str = "sandy brown",
    color_hum: str = "cornflower blue",
    color_hum_crys: str = "yellow green",
    color_mut_wt: str = "chocolate",
    color_mut_hum: str = "dark orchid",
    color_mut_wt_crys: str = "dark green",

):
    """
    Write a ChimeraX script (.cxc) that does the same as above using ChimeraX commands.
    Files need to be AHo aligned. 
    """
    lines = []
    lines.append(f'open "{os.path.basename(wt_pdb)}"')   # -> #1 in ChimeraX numbering
    lines.append(f'open "{os.path.basename(hum_pdb)}"')  # -> #2
    lines.append(f"color #1 {color_wt}")
    lines.append(f"color #2 {color_hum}")

    if wt_pdb_crys is not None: 
        lines.append(f'open "{os.path.basename(wt_pdb_crys)}"') 
        lines.append(f"color #3 {color_hum_crys}")

    lines.append("cartoon")
    h_muts = get_mutated_positions_from_aho_renumbered_pdbs(wt_pdb, wt_chain_h, hum_pdb, hum_chain_h)
    h_mutated_positions_aho = [pos for (pos, wt_aa, hum_aa) in h_muts]

    for pos in h_mutated_positions_aho:
        # style sticks; restrict to sidechain atoms with 'sidechain' selector
        lines.append(f"select #1/{wt_chain_h}:{pos}")
        lines.append(f"show sel atoms")
        lines.append(f"color sel {color_mut_wt}")
        lines.append(f"color sel byhetero")

        lines.append(f"select #2/{hum_chain_h}:{pos}")
        lines.append(f"show sel atoms")
        lines.append(f"color sel {color_mut_hum}")
        lines.append(f"color sel byhetero")

        if wt_pdb_crys is not None:
            lines.append(f"select #3/{wt_chain_h_crys}:{pos}")
            lines.append(f"show sel atoms")
            lines.append(f"color sel {color_mut_wt_crys}")
            lines.append(f"color sel byhetero")

    if not is_vhh:
        l_muts = get_mutated_positions_from_aho_renumbered_pdbs(wt_pdb, wt_chain_l, hum_pdb, hum_chain_l)
        l_mutated_positions_aho = [pos for (pos, wt_aa, hum_aa) in l_muts]

        for pos in l_mutated_positions_aho:
            # style sticks; restrict to sidechain atoms with 'sidechain' selector
            lines.append(f"select #1/{wt_chain_l}:{pos}")
            lines.append(f"show sel atoms")
            lines.append(f"color sel {color_mut_wt}")
            lines.append(f"color sel byhetero")

            lines.append(f"select #2/{hum_chain_l}:{pos}")
            lines.append(f"show sel atoms")
            lines.append(f"color sel {color_mut_hum}")
            lines.append(f"color sel byhetero")

            if wt_pdb_crys is not None:
                lines.append(f"select #3/{wt_chain_l_crys}:{pos}")
                lines.append(f"show sel atoms")
                lines.append(f"color sel {color_mut_wt_crys}")
                lines.append(f"color sel byhetero")
                

    lines.append(f"select H")
    lines.append(f"hide sel target a")
    lines.append(f"select clear")
    lines.append(f"set bgColor white")
    lines.append(f"graphics silhouettes true")

    lines.append("view")
    with open(out_cxc_path, "w") as f:
        f.write("\n".join(lines))



# (c) 2023 Sormannilab and Aubin Ramon
#
# Humanisation pipelines of nanobodies of AbNatiV VH and VHH assesments
#
# ============================================================================

import os
import numpy as np

from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq

from ..model.alignment.mybio import print_Alignment_pap
from ..model.alignment.mybio import renumber_Fv_pdb_file
from ..model.scoring_functions import abnativ_scoring, fr_aho_indices

from .struct import compute_cdr_displacement_between_two_structures
from .chimerax_visualition import write_chimerax_cxc_visualisation
from .humanisation_utils import humanise_enhanced_sampling, humanise_exhaustive_sampling, predict_struct_vhh


def abnativ_vhh_humanisation(wt_seq: str, name_seq: str, nat_vh:str = 'VH', nat_vhh:str = 'VHH', output_dir: str='abnativ_humanisation', pdb_file: str=None, ch_id: str=None, 
                             allowed_user_positions:list=fr_aho_indices, is_brute: bool=False, threshold_abnativ_score:float=.98,
                             threshold_rasa_score:float=0.15, perc_allowed_decrease_vhh:float=2e-2, forbidden_mut:list=['C','M'],
                             a:float=.8,b:float=.2, seq_ref: str=None, name_seq_ref: str=None, verbose: bool=True) -> None: 
    '''Run AbNatiV humanisation pipeline on a nanobody sequence with a dual-control strategy 
    that aims to increase the AbNatiV VH-hummanness of a sequence
    while retaining its VHH-nativeness.

    Two sampling methods are available:
        - enhanced sampling (is_brute=False): iteratively explores
            the mutational space aiming for rapid convergence to generate a single humanised sequence.
        - exhaustive sampling (is_brute=True): It assesses all mutation combinations within 
            the available mutational space (PSSM-allowed mutations) and selects the best sequences (Pareto Front).

    See parameters for further details. 
    
    Parameters
    ----------
        - wt_seq: str
            Unaligned sequence string 
        - name_seq: str
        - nat_vh: str
            VH model type to compute the VH-nativeness on ['VH', 'VH2']
        - nat_vhh: str
            VHH model type to compute the VHH-nativeness on ['VHH', 'VHH2']
        - output_dir:str
            Directory where to save files
        - pdb_file: str
            Filepath to the pdb file of the wt_seq
        - ch_id: str
            Chain id of the pfb pdb_file
        - allowed_user_aho_positions: list of int
            List of AHo positions allowed by the user to make mutation on (default: framework positions)
        - is_brute: bool
            If True, runs brute method rather than the enhanced sampling method
        - threshold_abnativ_score: float 
            Bellow the AbNatiV VH threshold score, a position is considered as a liability
        - threshold_rasa_score: float 
            Above this threshold, the residue is considered solvent exposed and is considered for mutation
        - perc_allowed_decrease_vhh: float
            Maximun ΔVHH score decrease allowed for a mutation
        - forbidden_mut: list
            List of residues to ban for mutation i.e. ['C','M']
        - a: float
            Used in multi-objective selection function: aΔVH+bΔVHH
        - b: float
            Used in multi-objective selection function: aΔVH+bΔVHH
        - seq_ref: str 
            If None, does not plot any references in the profiles. If str, will plot it
        - name_seq_ref: str
        - verbose: bool
            
    Returns
    -------
        - dict with keys will be AHo numbers, values will be allowed substitutions 
            according to the PSSM criteria '''

    # Create folders
    if not os.path.exists(output_dir): os.makedirs(output_dir)

    seq_dir = os.path.join(output_dir, name_seq)
    if not os.path.exists(seq_dir): os.makedirs(seq_dir)

    pdb_dir = os.path.join(seq_dir, 'structures')
    if not os.path.exists(pdb_dir): os.makedirs(pdb_dir)

    # Aligned WT sequence to take into account additional terminal residues added during the alignment process
    seq_records = [SeqRecord(Seq(wt_seq), id='single_seq')]
    wt_vh_seq_abnativ_df, wt_vh_profile_abnativ_df = abnativ_scoring(nat_vhh, seq_records, batch_size=1,mean_score_only=False, 
                                                                    do_align=True, is_VHH=True, verbose=False)
    al_wt_seq = ''.join(wt_vh_seq_abnativ_df['aligned_seq'])

    if '2' in nat_vh: is_abnativ2=True
    else: is_abnativ2=False

    # If provided, renumber PDB file into AHo numbering to compute RASA on
    if pdb_file is not None:
        aho_pdb_file = os.path.join(pdb_dir, os.path.basename(pdb_file).split('.')[0] + '_aho.pdb')
        try: 
            renumber_Fv_pdb_file(pdb_file, ch_id, None, is_VHH=True, scheme='AHo', outfilename=aho_pdb_file, check_AHo_CDR_gaps=is_abnativ2)
        except ValueError: 
            print('Difficulties to read PDB, please make sure it is as cleaned as possible or use the structure prediction option\
                    by writing "None" in the pdb_file option')
        pdb_file = aho_pdb_file

    # Run AbNatiV humanisation
    id_wt = name_seq + '_abnativ_wt'
    seq_records = list()
    seq_records.append(SeqRecord(Seq(al_wt_seq.replace('-','')), id=id_wt))
    
    if not is_brute: 
        if verbose: print(f'\n>> ENHANCED SAMPLING <<\n\n## {name_seq} ##\n')
        seq_hum =  humanise_enhanced_sampling(al_wt_seq.replace('-',''), name_seq, nat_vh, True, pdb_file, ch_id, seq_dir, allowed_user_positions,
                                        threshold_abnativ_score, threshold_rasa_score, perc_allowed_decrease_vhh=perc_allowed_decrease_vhh, forbidden_mut=forbidden_mut,
                                        a=a, b=b, seq_ref=seq_ref, name_seq_ref=name_seq_ref, verbose=verbose, nat_vhh=nat_vhh, pdb_dir=pdb_dir)
    
        id_hum = name_seq + '_abnativ_hum_enhanced'
        

    else: 
        if verbose: print(f'\n>> EXHAUSTIVE SAMPLING <<\n\n## {name_seq} ##\n')
        df_pareto_exhaustive = humanise_exhaustive_sampling(al_wt_seq.replace('-',''), nat_vh, True, name_seq, pdb_file, ch_id, seq_dir, pdb_dir, allowed_user_positions,
                                    threshold_abnativ_score, threshold_rasa_score, nat_vhh, perc_allowed_decrease_vhh, forbidden_mut,
                                    seq_ref, name_seq_ref)
        
        # Pre-process done on the higher number of mutations
        idx_best = df_pareto_exhaustive['count_mut'].idxmax()
        seq_hum, max_count = df_pareto_exhaustive.loc[idx_best, 'aligned_seq'].replace('-', ''), df_pareto_exhaustive.loc[idx_best, 'count_mut']
        id_hum = name_seq + f'_abnativ_hum_exhaustive_{max_count}'

    ### Pre-process the humanised results ###
    seq_records.append(SeqRecord(Seq(seq_hum), id=id_hum))

    # Final AbNatiV score and save
    vh_abnativ_df_mean, _ = abnativ_scoring(nat_vh, seq_records,  batch_size=1, mean_score_only=False, do_align=True, is_VHH=True)
    vhh_abnativ_df_mean, _ = abnativ_scoring(nat_vhh, seq_records, batch_size=1, mean_score_only=False, do_align=True, is_VHH=True)
    merge_abnativ_df = vh_abnativ_df_mean.merge(vhh_abnativ_df_mean, how='inner')

    # Print pap file 
    dict_all_seqs = {id_wt:wt_seq, id_hum:seq_hum}
    print_Alignment_pap(dict_all_seqs, os.path.join(seq_dir,f'{id_hum}.pap'), nchar_id=18)

    # Model WT/HUM structures
    wt_pred_pdb_file, wt_pred_ch_id, _, _, _, _ = predict_struct_vhh(al_wt_seq.replace('-',''), id_wt, pdb_dir, is_abnativ2=is_abnativ2)
    hum_pred_pdb_file, hum_pred_ch_id, _, _, _, _ = predict_struct_vhh(seq_hum, id_hum, pdb_dir, is_abnativ2=is_abnativ2)
    
    # Compute CDR-RMSD between the WT model and the humanised model
    fp_aligned_wt_model, scaffold, h_cdr1, h_cdr2, h_cdr3 = compute_cdr_displacement_between_two_structures(id_hum, hum_pred_pdb_file, hum_pred_ch_id, None, False,
                                                                id_wt, wt_pred_pdb_file, wt_pred_ch_id, None, False,
                                                                is_vhh=True, is_abnativ2=is_abnativ2, o_dir_pdb_models=pdb_dir)
                                                                
    
    merge_abnativ_df['RMSD (A)\nwith modelled WT\nof scaffold\nupon scaffold-superimposition'] = [np.nan, scaffold]
    merge_abnativ_df['RMSD (A)\nwith modelled WT\nof H-CDR1\nupon scaffold-superimposition'] = [np.nan, h_cdr1]
    merge_abnativ_df['RMSD (A)\nwith modelled WT\nof H-CDR2\nupon scaffold-superimposition'] = [np.nan, h_cdr2]
    merge_abnativ_df['RMSD (A)\nwith modelled WT\nof H-CDR3\nupon scaffold-superimposition'] = [np.nan, h_cdr3]

    # Compute CDR-RMSD between the WT crystal and the humanised model
    if pdb_file is not None:
        fp_aligned_wt_crys, scaffold_cryst, h_cdr1_cryst, h_cdr2_cryst, h_cdr3_cryst = compute_cdr_displacement_between_two_structures(id_hum, hum_pred_pdb_file, hum_pred_ch_id, None, False,
                                                                                                id_wt + '_cryst', pdb_file, ch_id, None, False,
                                                                                                is_vhh=True, is_abnativ2=is_abnativ2, o_dir_pdb_models=pdb_dir)

        merge_abnativ_df['RMSD (A)\nwith crystal WT\nof scaffold\nupon scaffold-superimposition'] = [np.nan, scaffold_cryst]
        merge_abnativ_df['RMSD (A)\nwith crystal WT\nof H-CDR1\nupon scaffold-superimposition'] = [np.nan, h_cdr1_cryst]
        merge_abnativ_df['RMSD (A)\nwith crystal WT\nof H-CDR2\nupon scaffold-superimposition'] = [np.nan, h_cdr2_cryst]
        merge_abnativ_df['RMSD (A)\nwith crystal WT\nof H-CDR3\nupon scaffold-superimposition'] = [np.nan, h_cdr3_cryst]
    else:
        fp_aligned_wt_crys = None

    merge_abnativ_df.to_csv(os.path.join(seq_dir, f'{id_hum}.csv'))

    # Print on chimera mutated residues 
    fp_chimerax_cxc = os.path.join(pdb_dir, f"{id_hum}_mut_chimerax.cxc")
    write_chimerax_cxc_visualisation(fp_aligned_wt_model, wt_pred_ch_id, None, 
                                        hum_pred_pdb_file, hum_pred_ch_id, None,
                                        fp_aligned_wt_crys, ch_id, None, # Crystal structure if present
                                        is_vhh=True, out_cxc_path=fp_chimerax_cxc)
    
    return merge_abnativ_df


# (c) 2023 Sormannilab and Aubin Ramon
#
# Training of the AbNatiV model with the Pytorch Lightning module.
#
# ============================================================================

from .model.abnativ import AbNatiV_Model
from .model.onehotencoder import data_loader_masking_bert_onehot_fasta
from .model.alignment.mybio import anarci_alignments_of_Fv_sequences

import random
import os
import argparse
import yaml
from Bio import SeqIO

import pytorch_lightning as pl
from pytorch_lightning.loggers import MLFlowLogger, WandbLogger
from pytorch_lightning.callbacks import ModelCheckpoint


def run(args: argparse.Namespace):
    # fix seed for reproductibility
    pl.seed_everything(11)
    random.seed(11)

    ## BUILDING MODULES ##
    with open(args.hparams, "r") as f:
        hparams = yaml.safe_load(f)

    model = AbNatiV_Model(hparams)

    ## ALIGNMENT ##
    if args.do_align:
        # Train
        print(f"\n### ANARCI alignment of {args.train_filepath}###\n")
        VH, VK, VL, failed, mischtype = anarci_alignments_of_Fv_sequences(
            args.train_filepath,
            isVHH=args.is_VHH,
            del_cyst_misalign=True,
            run_parallel=args.ncpu,
        )
        VH.add(VK)
        VH.add(VL)
        fp_al_train = "temp_al_train_seqs.fa"
        VH.Print(fp_al_train)

        # Val
        print(f"\n### ANARCI alignment of {args.val_filepath}###\n")
        VH, VK, VL, failed, mischtype = anarci_alignments_of_Fv_sequences(
            args.val_filepath,
            isVHH=args.is_VHH,
            del_cyst_misalign=True,
            run_parallel=args.ncpu,
        )
        VH.add(VK)
        VH.add(VL)
        fp_al_val = "temp_al_val_seqs.fa"
        VH.Print(fp_al_val)

    else:
        fp_al_train = args.train_filepath
        fp_al_val = args.val_filepath

    ## DATA LOADING ##
    train_loader = data_loader_masking_bert_onehot_fasta(
        fp_al_train,
        hparams["batch_size"],
        hparams["perc_masked_residues"],
        is_masking=True,
    )

    val_loader = data_loader_masking_bert_onehot_fasta(
        fp_al_val, hparams["batch_size"], perc_masked_residues=0, is_masking=False
    )

    ## TRAINING ##
    # Logging
    logger = WandbLogger(
        project=args.model_name, name=hparams["run_name"], log_model=False
    )

    # Checkpointing
    ckpt_root_dir = os.path.join("checkpoints", hparams["run_name"])
    ckpt_callback = ModelCheckpoint(ckpt_root_dir, save_top_k=-1)  # to save every epoch

    trainer = pl.Trainer(
        limit_train_batches=1,
        limit_val_batches=hparams["limit_val_batches"],
        max_epochs=hparams["max_epochs"],
        deterministic=True,
        accelerator="auto",
        logger=logger,
        callbacks=[ckpt_callback],
    )

    # Training
    trainer.fit(model, train_dataloaders=train_loader, val_dataloaders=val_loader)

    # Remove temporary files
    if args.do_align:
        os.remove(fp_al_val)
        os.remove(fp_al_val)



# (c) 2023 Sormannilab and Aubin Ramon
#
# Lightning testing of the AbNatiV model.
#
# ============================================================================

from .model.scoring_functions import abnativ_scoring, abnativ_scoring_csv
from .model.onehotencoder import alphabet
import argparse
import os

from .model.utils import is_protein
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq

from .init import PRETRAINED_MODELS_DIR

def run(args: argparse.Namespace):

    # Check that models are downloaded
    if not os.path.exists(PRETRAINED_MODELS_DIR):
        raise Exception("Models not found. Please run 'abnativ update' to download the models.")

    if not os.path.exists(args.output_directory):
        os.makedirs(args.output_directory)

    ## DATA SCORING ##
    batch_size = 128

    # Check if input is a CSV file
    is_csv_input = (
        os.path.isfile(args.input_filepath_or_seq) and
        args.input_filepath_or_seq.lower().endswith('.csv')
    )

    if is_csv_input:
        # Use CSV scoring with column preservation
        abnativ_df_mean, abnativ_df_profile = abnativ_scoring_csv(
            model_type=args.nativeness_type,
            input_csv=args.input_filepath_or_seq,
            col_id=args.col_id,
            col_seq=args.col_seq,
            batch_size=batch_size,
            chunk_size=args.chunk_size,
            mean_score_only=args.mean_score_only,
            do_align=args.do_align,
            is_VHH=args.is_VHH,
            is_plotting_profiles=args.is_plotting_profiles,
            output_dir=args.output_directory,
            output_id=args.output_id,
            verbose=True,
            run_parall_al=args.ncpu,
            preserve_columns=True,
        )

    else:
        # Original FASTA/sequence handling
        # Load the fasta file
        if os.path.isfile(args.input_filepath_or_seq):
            seq_records = list(SeqIO.parse(args.input_filepath_or_seq, 'fasta'))

        # If just a single sequence string
        elif is_protein(args.input_filepath_or_seq, alphabet + ['X']):
            seq_records = [SeqRecord(Seq(args.input_filepath_or_seq), id='single_seq')]

        else:
            raise Exception(f'Can not find the input file {args.input_filepath_or_seq} or if you gave a single protein sequence make sure it is composed only of {alphabet}')

        abnativ_df_mean, abnativ_df_profile = abnativ_scoring(
            args.nativeness_type, seq_records, batch_size, args.mean_score_only,
            args.do_align, args.is_VHH, args.is_plotting_profiles, args.output_directory, args.output_id,
            run_parall_al=args.ncpu
        )

    ## DATA SAVING ##
    print(f'\n-> Scores being saved in {args.output_directory}\n')
    abnativ_df_mean.to_csv(os.path.join(args.output_directory, f'{args.output_id}_abnativ_seq_scores.csv'), index=False)
    if not args.mean_score_only:
        abnativ_df_profile.to_csv(os.path.join(args.output_directory, f'{args.output_id}_abnativ_res_scores.csv'), index=False)
    if args.is_plotting_profiles:
        save_profile_fp = os.path.join(args.output_directory, f'{args.output_id}_profiles')
        print(f'\n-> Profile plots saved in {save_profile_fp}\n')



# (c) 2023 Sormannilab and Aubin Ramon
#
# AbNatiV model, Pytorch version
#
# ============================================================================
# fmt: off

import os
from typing import Tuple
import math
import pickle
import numpy as np
import pandas as pd
from collections import defaultdict
import logging
from datetime import datetime
from einops import rearrange
from scipy.stats import pearsonr
from sklearn.metrics import roc_curve, auc, precision_recall_curve
from sklearn.metrics import confusion_matrix
import time

# Configure logging
logging.basicConfig(level=logging.INFO, format='%(message)s')

from .vq import VectorQuantize
from .utils import find_optimal_cnn1d_padding, find_out_padding_cnn1d_transpose

import torch
from torch import nn 
from torch.nn import functional as F

from .alignment.aho_consensus import (
    cdr1_aho_indices,
    cdr2_aho_indices,
    cdr3_aho_indices,
    fr_aho_indices,
)

import pytorch_lightning as pl
from einops.layers.torch import Rearrange

# Perform real-time analysis during each validation epoch.
# Save the following metrics and scores:
# 1. Validation scores in 'val_scores.csv'
# 2. PSSM scores in 'pssm_scores.csv'
# 3. Accuracy, Loss, reconstruction error, and perplexity in a pickle file
# Additionally, plot the following:
# - Receiver Operating Characteristic (ROC) curves
# - Area Under the Curve (AUC) scores
from .analysis_functions import calculate_pssm_aucs_paired_with_mismatch, get_abnativ_scores_per_batch, evaluate_paired_pssm_with_mismatch, evaluate_pairing_pred_with_mismatch

def binary_focal_loss(probs: torch.Tensor, targets: torch.Tensor, gamma=2.0):
    """
    Focal loss for probabilities.
    """
    bce_loss = nn.BCELoss(reduction='none')
    loss = bce_loss(probs, targets)

    pt = torch.where(targets == 1, probs, 1 - probs)
    focal_weight = (1 - pt) ** gamma
    focal_loss = focal_weight * loss
    return focal_loss


class RotaryEncoding(nn.Module):
    """
    Rotary positional embeddings applied to queries and keys in a transformer.

    paper: RoFormer: Enhanced Transformer with Rotary Position Embedding
    link: https://arxiv.org/abs/2104.09864
    """

    def __init__(self, head_dim: int, theta: float = 10000.0):
        """
        :param head_dim: The per head embedding dimension of the query or key vectors.
        :param theta: Scaling factor for the rotary frequencies.
        """
        super().__init__()

        if head_dim % 2 != 0:
            raise ValueError(
                f"Rotary encoding requires an even embedding dimension, "
                f"but received head_dim={head_dim}. Please use an even number."
            )

        freqs = 1.0 / (theta ** (torch.arange(0, head_dim, 2) / head_dim))
        self.register_buffer("freqs", freqs)

    def forward(self, v_i: torch.Tensor) -> torch.Tensor:
        """
        Applies rotary positional encoding to the input vector.
        A 2D case is: (x, y) = [[cos(theta), -sin(theta)], [sin(theta), cos(theta)]] @ (x, y)

        :param v_i: Input tensor of shape :code:`(batch_size, num_heads, seq_len, head_dim)`, where head_dim should
            match `head_dim` in the constructor.
        :return: Tensor with rotary positional encoding applied, of the same shape as input.
        """
        _, _, seq_len, _ = v_i.shape

        t = torch.arange(seq_len, device=v_i.device, dtype=v_i.dtype).float()
        angles = torch.einsum("i,j->ij", t, self.freqs)
        emb_sin = torch.sin(angles).repeat_interleave(2, dim=-1)
        emb_cos = torch.cos(angles).repeat_interleave(2, dim=-1)

        v_rotated = (v_i * emb_cos) + (self.rotate_half(v_i) * emb_sin)
        return v_rotated

    @staticmethod
    def rotate_half(x: torch.Tensor) -> torch.Tensor:
        """
        Rotates the pairwise elements of the input tensor by 90 degrees.
        (x1, x2) -> (-x2, x1)

        Example:
        x = tensor([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        x = RotaryEncoding.rotate_half(x)
        x = tensor([-1,  0, -3,  2, -5,  4, -7,  6, -9,  8])
        """
        x = rearrange(x, "... (d r) -> ... d r", r=2)
        x1, x2 = x.unbind(dim=-1)
        x = torch.stack((-x2, x1), dim=-1)
        return rearrange(x, "... d r -> ... (d r)")


class GatedSelfAttention(nn.Module):
    """
    Gated self-attention layer with a rotary position embedding.
    """

    def __init__(
        self,
        embed_dim: int,
        num_heads: int,
        dropout_p: float = 0.1,
        rotary_theta: float = 10000.0,
    ):
        """
        :param embed_dim: The dimensionality of the input/output embeddings.
        :param num_heads: The number of attention heads to use. The embedding dimension :code:`embed_dim`
            must be divisible by this number.
        :param dropout_p: Dropout probability for the attention weights.
        :param rotary_theta: Theta parameter for the rotary positional encoding.
        """
        super().__init__()

        if embed_dim % num_heads != 0:
            raise ValueError(
                f"Embedding dimension {embed_dim} should be divisible by the number of heads {num_heads}."
            )

        self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.dropout_p = dropout_p

        self.layer_norm = nn.LayerNorm(embed_dim)

        self.linear_q = nn.Linear(embed_dim, embed_dim, bias=False)
        self.linear_k = nn.Linear(embed_dim, embed_dim, bias=False)
        self.linear_v = nn.Linear(embed_dim, embed_dim, bias=False)

        self.linear_no_bias_g = nn.Linear(embed_dim, embed_dim, bias=False)
        self.out = nn.Sequential(
            nn.Linear(embed_dim, embed_dim, bias=False), nn.Dropout(dropout_p)
        )

        self.sigmoid = nn.Sigmoid()

        head_dim = embed_dim // num_heads
        self.rotary_encoding = RotaryEncoding(head_dim=head_dim, theta=rotary_theta)

    def forward(self, s_i: torch.Tensor) -> torch.Tensor:
        """
        Forward pass on a single tensor of embeddings.

        :param s_i: :code:`(batch_size, seq_len, embed_dim)` tensor of embeddings.
        :return: Updated tensor of embeddings, same shape as the input.
        """

        # Input projections
        s_i = self.layer_norm(s_i)

        q_i = rearrange(self.linear_q(s_i), "b i (h d) -> b h i d", h=self.num_heads)
        k_i = rearrange(self.linear_k(s_i), "b i (h d) -> b h i d", h=self.num_heads)
        v_i = rearrange(self.linear_v(s_i), "b i (h d) -> b h i d", h=self.num_heads)
        g_i = rearrange(
            self.sigmoid(self.linear_no_bias_g(s_i)),
            "b i (h d) -> b h i d",
            h=self.num_heads,
        )

        # Rotary positional encoding
        q_i = self.rotary_encoding(q_i)
        k_i = self.rotary_encoding(k_i)

        # Attention
        attn_output = F.scaled_dot_product_attention(q_i, k_i, v_i) * g_i
        attn_output = rearrange(attn_output, "b h i d -> b i (h d)")

        # Output projections
        s_i = self.out(attn_output)
        return s_i


class SelfAttentionBlock(nn.Module):
    """A single block that performs self-attention followed by a SwiGLU transition layer."""

    def __init__(
        self,
        d_embedding: int,
        num_heads: int,
        d_ff: int,
        dropout_p: float = 0.1,
    ):
        """
        :param embed_dim: Dimension of the input embeddings.
        :param num_heads: Number of attention heads.
        :param dff: Dimension of the hiden layer in the SIWLU 
        :param dropout_p: Dropout probability for the attention weights.
        """
        super().__init__()

        self.attention = GatedSelfAttention(
            embed_dim=d_embedding,
            num_heads=num_heads,
            dropout_p=dropout_p,
        )

        self.linear_w = nn.Linear(d_embedding, d_ff)
        self.linear_v = nn.Linear(d_embedding, d_ff)
        self.d_emb_linear = nn.Linear(d_ff, d_embedding)
        self.swish = nn.SiLU()

        self.layernorm1 = nn.LayerNorm(d_embedding, eps=1e-6)
        self.layernorm2 = nn.LayerNorm(d_embedding, eps=1e-6)

        self.dropout = nn.Dropout(dropout_p)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass on a single tensor of embeddings.

        :param x: :code:`(batch_size, seq_len, embed_dim)` tensor of embeddings.
        :return: Tensor of updated embeddings, same shape as the input.
        """
        # Attention
        attn_output = self.attention(x)
        x = x + self.dropout(attn_output)
        x = self.layernorm1(x)

        # SWIGLU
        linear_output = self.d_emb_linear(self.swish(self.linear_w(x))*self.linear_v(x))
        x = x + self.dropout(linear_output)
        x = self.layernorm2(x) # (batch_size, input_seq_len, d_embedding) + residual 

        return x


class AdaptiveLayerNorm(nn.Module):
    """
    Implements adaptive layer normalisation similar to AlphaFold3 (algorithm 26),
    which performs layer normalisation on the primary tensor conditional on the
    secondary tensor.

    The key difference with the AlphaFold implementation is that the secondary tensor is
    aggregated via a mean over the token dimension before projecting, to allow for different
    numbers of tokens in the primary and secondary tensors.
    """

    def __init__(self, embed_dim: int):
        """
        :param embed_dim: The number of features in the input tensors.
        """
        super().__init__()

        self.ln_no_affine = nn.LayerNorm(embed_dim, elementwise_affine=False)
        self.ln_no_bias = nn.LayerNorm(embed_dim, bias=False)

        self.linear_s = nn.Linear(embed_dim, embed_dim)
        self.linear_no_bias_s = nn.Linear(embed_dim, embed_dim, bias=False)

        self.sigmoid = nn.Sigmoid()

    def forward(self, a: torch.Tensor, s: torch.Tensor) -> torch.Tensor:
        """
        :param a: The input tensor to apply layer norm on.
        :param s: A secondary tensor for adaptively modulating the normalization. It is assumed that the
            token dimension of this tensor is the second-to-last dimension.
        :return: The normalized tensor.
        """

        a = self.ln_no_affine(a)
        s = torch.mean(s, dim=-2, keepdim=True)
        s = self.ln_no_bias(s)
        a = self.sigmoid(self.linear_s(s)) * a + self.linear_no_bias_s(s)

        return a


class GatedCrossAttention(nn.Module):
    """
    Cross-attention layer for heavy/light chain embedding interactions. This layer uses
    an adaptive layer norm and a gating mechanism to modulate the attention weights obtained
    from the cross attention mechanism.
    """

    def __init__(
        self,
        embed_dim: int,
        num_heads: int,
        dropout_p: float = 0.1,
        gate_bias_value: float = 0.0,
    ):
        """
        :param embed_dim: The dimensionality of the input/output embeddings.
        :param num_heads: The number of attention heads to use. The embedding dimension :code:`embed_dim`
            must be divisible by this number.
        :param dropout_p: The dropout probability for the attention weights.
        :param gate_bias_value: The initialisation value for the gate bias. This can be initialised to a
            large negative value to prevent any cross-attention modulation at the start of training.
        """

        super().__init__()

        self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.dropout_p = dropout_p

        self.layer_norm_heavy = AdaptiveLayerNorm(embed_dim)
        self.layer_norm_light = AdaptiveLayerNorm(embed_dim)

        self.q_heavy = nn.Linear(embed_dim, embed_dim, bias=False)
        self.k_light = nn.Linear(embed_dim, embed_dim, bias=False)
        self.v_light = nn.Linear(embed_dim, embed_dim, bias=False)

        self.q_light = nn.Linear(embed_dim, embed_dim, bias=False)
        self.k_heavy = nn.Linear(embed_dim, embed_dim, bias=False)
        self.v_heavy = nn.Linear(embed_dim, embed_dim, bias=False)

        self.linear_g_heavy = nn.Linear(embed_dim, embed_dim)
        self.linear_g_light = nn.Linear(embed_dim, embed_dim)

        self.linear_g_heavy.bias.data.fill_(gate_bias_value)
        self.linear_g_light.bias.data.fill_(gate_bias_value)

        self.sigmoid = nn.Sigmoid()

        self.output_heavy = nn.Sequential(
            nn.Linear(embed_dim, embed_dim), nn.Dropout(dropout_p)
        )
        self.output_light = nn.Sequential(
            nn.Linear(embed_dim, embed_dim), nn.Dropout(dropout_p)
        )

    def forward(
        self, heavy_repr: torch.Tensor, light_repr: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Forward pass on a pair of heavy and light chain embeddings.

        :param heavy_repr: :code:`(batch_size, vh_seq_len, embed_dim)` tensor of heavy chain embeddings.
        :param light_repr: :code:`(batch_size, vl_seq_len, embed_dim)` tensor of light chain embeddings.
        :return: Tuple of updated heavy and light chain embeddings, same shape as the inputs.
        """

        heavy_repr = self.layer_norm_heavy(heavy_repr, light_repr)
        light_repr = self.layer_norm_light(light_repr, heavy_repr)

        # Heavy to Light cross-attention
        q_heavy = rearrange(
            self.q_heavy(heavy_repr), "b i (h d) -> b h i d", h=self.num_heads
        )
        k_light = rearrange(
            self.k_light(light_repr), "b i (h d) -> b h i d", h=self.num_heads
        )
        v_light = rearrange(
            self.v_light(light_repr), "b i (h d) -> b h i d", h=self.num_heads
        )
        g_heavy = rearrange(
            self.sigmoid(self.linear_g_heavy(heavy_repr)),
            "b i (h d) -> b h i d",
            h=self.num_heads,
        )

        attn_output_heavy = (
            F.scaled_dot_product_attention(q_heavy, k_light, v_light) * g_heavy
        )
        attended_heavy = rearrange(attn_output_heavy, "b h i d -> b i (h d)")

        # Light to Heavy cross-attention
        q_light = rearrange(
            self.q_light(light_repr), "b i (h d) -> b h i d", h=self.num_heads
        )
        k_heavy = rearrange(
            self.k_heavy(heavy_repr), "b i (h d) -> b h i d", h=self.num_heads
        )
        v_heavy = rearrange(
            self.v_heavy(heavy_repr), "b i (h d) -> b h i d", h=self.num_heads
        )
        g_light = rearrange(
            self.sigmoid(self.linear_g_light(light_repr)),
            "b i (h d) -> b h i d",
            h=self.num_heads,
        )

        attn_output_light = (
            F.scaled_dot_product_attention(q_light, k_heavy, v_heavy) * g_light
        )
        attended_light = rearrange(attn_output_light, "b h i d -> b i (h d)")

        # Output projections
        updated_heavy = self.output_heavy(attended_heavy)
        updated_light = self.output_light(attended_light)

        return updated_heavy, updated_light
    

class SwiGLU(nn.Module):
    """A swish-gated linear unit (SwiGLU)."""

    def __init__(self, input_dim: int, output_dim: int, bias: bool = False):
        """
        :param input_dim: The dimension of the input tensor.
        :param output_dim: The dimension of the output tensor.
        """
        super().__init__()

        self.linear_1 = nn.Linear(input_dim, output_dim, bias=bias)
        self.linear_2 = nn.Linear(input_dim, output_dim, bias=bias)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass on a single tensor of embeddings.

        :param x: :code:`(..., input_dim)` tensor of embeddings.
        :return: :code:`(..., output_dim)` tensor of updated embeddings.
        """
        a = self.linear_1(x)
        b = self.linear_2(x)
        x = F.silu(a) * b

        return x


class ConditionedSwiGLUTransition(nn.Module):
    """
    A conditional SwiGLU-activated transition block. This implements Algorithm 25 from AlphaFold3,
    with the addition of dropout.
    """

    def __init__(
        self,
        embed_dim: int,
        hidden_dim: int,
        dropout_p: float = 0.1,
        gate_bias_value: float = -2.0,
    ):
        """
        :param embed_dim: Dimension of the input embeddings.
        :param dropout_p: Dropout probability for the transition layer.
        :param gate_bias_value: The initialisation value for the gate bias. This can be initialised to a
            large negative value to prevent any cross-attention modulation at the start of training.
        """
        super().__init__()

        self.layers = nn.Sequential(
            SwiGLU(embed_dim, hidden_dim),
            nn.Dropout(dropout_p),
            nn.Linear(hidden_dim, embed_dim),
        )

        self.norm = AdaptiveLayerNorm(embed_dim)

        self.linear_gate = nn.Linear(embed_dim, embed_dim)
        self.linear_gate.bias.data.fill_(gate_bias_value)

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """
        Forward pass to transform an input tensor :code:`x` conditional on a secondary tensor :code:`y`.

        :param x: Input tensor to be transformed.
        :param y: Tensor containing conditioning information.
        :return: The transformed tensor.
        """
        x = self.norm(x,y)
        out = self.layers(x)
        gate = torch.sigmoid(self.linear_gate(y))

        return out * gate


class CrossAttentionBlock(nn.Module):
  '''CrossAttention Transformer Block'''
  def __init__(self, d_embedding, num_heads, d_ff, dropout, 
               gate_bias_value: float = -2.0):
    super(CrossAttentionBlock, self).__init__()

    self.cross_attention = GatedCrossAttention(
            embed_dim=d_embedding,
            num_heads=num_heads,
            dropout_p=dropout,
            gate_bias_value=gate_bias_value,
        )

    self.dropout = nn.Dropout(dropout)

    self.transition_heavy = ConditionedSwiGLUTransition(
            embed_dim=d_embedding, hidden_dim=d_ff,
            dropout_p=dropout, gate_bias_value=gate_bias_value
        )
    
    self.transition_light = ConditionedSwiGLUTransition(
            embed_dim=d_embedding, hidden_dim=d_ff,
            dropout_p=dropout, gate_bias_value=gate_bias_value
        )

  def forward(self, vh_x, vl_x) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Args:
      vh_x: Tensor, shape [batch_size, l_red, d_embedding]
      vh_y: Tensor, shape [batch_size, l_red, d_embedding]
    """

    # Attention
    vh_output, vl_output = self.cross_attention(vh_x, vl_x)  # (batch_size, l_red, d_embedding)
    vh_x = vh_x + self.dropout(vh_output)
    vl_x = vl_x + self.dropout(vl_output)

    # SWIGLU
    vh_x = vh_x + self.transition_heavy(vh_x, vl_x) # (batch_size, input_seq_len, d_embedding) + residual 
    vl_x = vl_x + self.transition_light(vl_x, vh_x)

    return vh_x, vl_x


class Encoder(nn.Module):
  def __init__(self, d_embedding, kernel, stride, num_heads, num_mha_layers, d_ff,
              length_seq, alphabet_size, dropout=0):
    super(Encoder, self).__init__()

    self.num_mha_layers = num_mha_layers
    self.d_embedding = d_embedding

    # CNN1d embedding
    self.l_red, self.padding = find_optimal_cnn1d_padding(L_in=length_seq, K=kernel, S=stride)
    self.cnn_embedding =  nn.Sequential(Rearrange('b l r -> b r l'),
                nn.Conv1d(alphabet_size, d_embedding, kernel_size=kernel, stride=stride, padding=self.padding),
                Rearrange('b r l -> b l r'))

    # Positional encoding
    self.en_dropout = nn.Dropout(dropout)

    # MHA blocks
    self.en_MHA_blocks = nn.ModuleList([SelfAttentionBlock(d_embedding, num_heads, d_ff, dropout)
                       for _ in range(num_mha_layers)])

  def forward(self, x) -> torch.Tensor: 
    """
    Args:
      x: Tensor, shape [batch_size, input_seq_len, alphabet_size]
    """
    # CNN1d Embedding
    h = self.cnn_embedding(x) # (batch_size, l_red, d_embedding)
    h = self.en_dropout(h) 

    # MHA blocks
    for i, l in enumerate(self.en_MHA_blocks):
      h = self.en_MHA_blocks[i](h) # (batch_size, l_red, d_embedding)
    
    return h

  

class PairedEncoder(nn.Module):
  '''A paired encoder module with cross-attention between two 
  transformers to capture cross-correlation between the heavy and light chain.
  
  The paired encoder does not CNN1 embed anymore. It takes as inputs the output
  of the individual encoders (so post CNN1)'''

  def __init__(self, d_embedding, num_heads, d_ff,
              heavy_encoder: Encoder, light_encoder: Encoder, 
              freeze_heavy: bool=False, freeze_light: bool=False, 
              dropout=0, gate_bias_value:float=-2.0):
    
    super(PairedEncoder, self).__init__()

    if heavy_encoder.num_mha_layers != light_encoder.num_mha_layers:
            raise ValueError(
                "Heavy and light networks must have the same number of layers."
            )
    
    self.num_layers = heavy_encoder.num_mha_layers
    self.heavy_network = heavy_encoder
    self.light_network = light_encoder

    if freeze_heavy:
       for param in self.heavy_network.parameters():
          param.requires_grad = False
    if freeze_light:
       for param in self.light_network.parameters():
          param.requires_grad = False

    # Projections before cross attention
    self.vh_pre_projections = nn.ModuleList(
        [
            nn.Linear(heavy_encoder.d_embedding, d_embedding, bias=False)
            for _ in range(self.num_layers)
        ]
    )
    self.vl_pre_projections = nn.ModuleList(
        [
            nn.Linear(light_encoder.d_embedding, d_embedding, bias=False)
            for _ in range(self.num_layers)
        ]
      )

    # MHA blocks
    self.cross_attention_blocks = nn.ModuleList([CrossAttentionBlock(d_embedding, num_heads, d_ff, dropout, gate_bias_value)
                       for _ in range(self.num_layers)])
    
    
    # Projections after cross attention
    self.vh_post_projections = nn.ModuleList(
        [
            nn.Linear(d_embedding, heavy_encoder.d_embedding, bias=False)
            for _ in range(self.num_layers)
        ]
    )
    self.vl_post_projections = nn.ModuleList(
        [
            nn.Linear(d_embedding, light_encoder.d_embedding, bias=False)
            for _ in range(self.num_layers)
        ]
    )

  def forward(self, vh_x, vl_x) -> torch.Tensor: 
    """
    Args:
      vh_x: Tensor, shape [batch_size, input_seq_len, alphabet_size] representing heavy sequence
      vl_x: Tensor, shape [batch_size, input_seq_len, alphabet_size] representing light sequence
    """

    # Embedding
    vh_repr = self.heavy_network.cnn_embedding(vh_x) 
    vh_repr = self.heavy_network.en_dropout(vh_repr) 

    vl_repr = self.light_network.cnn_embedding(vl_x) 
    vl_repr = self.light_network.en_dropout(vl_repr) 

    # MHA blocks
    for i in range(self.num_layers):
        # Self-attention on each chain
        vh_repr = self.heavy_network.en_MHA_blocks[i](vh_repr)
        vl_repr = self.light_network.en_MHA_blocks[i](vl_repr)

        # Project to shared embedding space
        vh_repr_cross = self.vh_pre_projections[i](vh_repr)
        vl_repr_cross = self.vl_pre_projections[i](vl_repr)

        # Cross-attention between heavy and light representations
        vh_repr_cross, vl_repr_cross = self.cross_attention_blocks[i](
            vh_repr_cross, vl_repr_cross
        )

        # Project back to original embedding spaces, with residual connection
        vh_repr = vh_repr + self.vh_post_projections[i](vh_repr_cross)
        vl_repr = vl_repr + self.vl_post_projections[i](vl_repr_cross)
    
    return vh_repr, vl_repr



class Decoder(nn.Module):
  def __init__(self, d_embedding, kernel, stride, num_heads, num_mha_layers, d_ff,
                  length_seq, alphabet_size, dropout=0):
    super(Decoder, self).__init__()

    self.num_mha_layers = num_mha_layers
    self.d_embedding = d_embedding

    # Positional encoding
    self.l_red, self.padding = find_optimal_cnn1d_padding(L_in=length_seq, K=kernel, S=stride)
    self.de_dropout = nn.Dropout(dropout)

    # MHA blocks
    self.de_MHA_blocks = nn.ModuleList([SelfAttentionBlock(d_embedding, num_heads, d_ff, dropout)
                       for _ in range(num_mha_layers)])

    # Dense reconstruction
    self.dense_to_alphabet = nn.Linear(d_embedding, alphabet_size)
    self.dense_reconstruction = nn.Linear(alphabet_size*self.l_red, length_seq*alphabet_size)

    # CNN1d reconstruction
    self.out_pad = find_out_padding_cnn1d_transpose(L_obj=length_seq, L_in=self.l_red, K=kernel, S=stride, P=self.padding)
    self.cnn_reconstruction =  nn.Sequential(Rearrange('b l r -> b r l'),
                nn.ConvTranspose1d(d_embedding, alphabet_size, kernel_size=kernel, stride=stride, 
                              padding=self.padding, output_padding=self.out_pad),
                Rearrange('b r l -> b l r'))
    
  
  def forward(self, q) -> torch.Tensor:
    """
    Args:
      q: Tensor, shape [batch_size, l_red, d_embedding]
    """
    z = self.de_dropout(q) 

    # MHA blocks
    for i, l in enumerate(self.de_MHA_blocks):
      z = self.de_MHA_blocks[i](z) # (batch_size, l_red, d_embedding)
      
    # CNN reconstruction 
    z = self.cnn_reconstruction(z) # (batch_size, input_seq_len, alphabet_size)
    z_recon = F.softmax(z, dim=-1)

    return z_recon




class PairedDecoder(nn.Module):
  '''A paired decoder module with cross-attention between two 
  transformers to capture cross-correlation between the heavy and light chain.
  
  The paired decoder does not CNN1 embed anymore. It takes as inputs the output
  of the individual decoders (so post CNN1)'''

  def __init__(self, d_embedding, num_heads, d_ff,
              heavy_decoder: Decoder, light_decoder: Decoder, 
              freeze_heavy: bool=False, freeze_light: bool=False, 
              dropout=0, gate_bias_value:float=-2.0):
    
    super(PairedDecoder, self).__init__()

    if heavy_decoder.num_mha_layers != light_decoder.num_mha_layers:
            raise ValueError(
                "Heavy and light networks must have the same number of layers."
            )
    
    self.num_layers = heavy_decoder.num_mha_layers
    self.heavy_network = heavy_decoder
    self.light_network = light_decoder

    if freeze_heavy:
       for param in self.heavy_network.parameters():
          param.requires_grad = False
    if freeze_light:
       for param in self.light_network.parameters():
          param.requires_grad = False

    # Projections before cross attention
    self.vh_pre_projections = nn.ModuleList(
        [
            nn.Linear(heavy_decoder.d_embedding, d_embedding, bias=False)
            for _ in range(self.num_layers)
        ]
    )
    self.vl_pre_projections = nn.ModuleList(
        [
            nn.Linear(light_decoder.d_embedding, d_embedding, bias=False)
            for _ in range(self.num_layers)
        ]
      )

    # MHA blocks
    self.cross_attention_blocks = nn.ModuleList([CrossAttentionBlock(d_embedding, num_heads, d_ff, dropout, gate_bias_value)
                       for _ in range(self.num_layers)])
    
    
    # Projections after cross attention
    self.vh_post_projections = nn.ModuleList(
        [
            nn.Linear(d_embedding, heavy_decoder.d_embedding, bias=False)
            for _ in range(self.num_layers)
        ]
    )
    self.vl_post_projections = nn.ModuleList(
        [
            nn.Linear(d_embedding, light_decoder.d_embedding, bias=False)
            for _ in range(self.num_layers)
        ]
    )

  def forward(self, vh_x, vl_x) -> torch.Tensor: 
    """
    Args:
      vh_x: Tensor, shape [batch_size, input_seq_len, alphabet_size] representing heavy sequence
      vl_x: Tensor, shape [batch_size, input_seq_len, alphabet_size] representing light sequence
    """

    # Embedding
    vh_repr = self.heavy_network.de_dropout(vh_x)
    vl_repr = self.light_network.de_dropout(vl_x) 

    # MHA blocks
    for i in range(self.num_layers):
        # Self-attention on each chain
        vh_repr = self.heavy_network.de_MHA_blocks[i](vh_repr)
        vl_repr = self.light_network.de_MHA_blocks[i](vl_repr)

        # Project to shared embedding space
        vh_repr_cross = self.vh_pre_projections[i](vh_repr)
        vl_repr_cross = self.vl_pre_projections[i](vl_repr)

        # Cross-attention between heavy and light representations
        vh_repr_cross, vl_repr_cross = self.cross_attention_blocks[i](
            vh_repr_cross, vl_repr_cross
        )

        # Project back to original embedding spaces, with residual connection
        vh_repr = vh_repr + self.vh_post_projections[i](vh_repr_cross)
        vl_repr = vl_repr + self.vl_post_projections[i](vl_repr_cross)

    vh_pre_poj = vh_repr.clone()
    vl_pre_poj = vl_repr.clone()

    # CNN reconstruction 
    vh_repr = self.heavy_network.cnn_reconstruction(vh_repr) # (batch_size, input_seq_len, alphabet_size)
    vh_repr = F.softmax(vh_repr, dim=-1)

    vl_repr = self.light_network.cnn_reconstruction(vl_repr) # (batch_size, input_seq_len, alphabet_size)
    vl_repr = F.softmax(vl_repr, dim=-1)
    
    return vh_repr, vl_repr, vh_pre_poj, vl_pre_poj




class AbNatiV_Paired_Model(pl.LightningModule):
  def __init__(self, hparams: dict):
    super(AbNatiV_Paired_Model, self).__init__()

    # MODEL
    self.run_name = hparams["run_name"]

    # Paired encoder
    vh_encoder = Encoder(hparams['d_embedding'], hparams['kernel'], hparams['stride'], hparams['num_heads'], 
                            hparams['num_mha_layers'], hparams['d_ff'], hparams['length_seq'], 
                            hparams['alphabet_size'], dropout=hparams['drop'])
    
    vl_encoder = Encoder(hparams['d_embedding'], hparams['kernel'], hparams['stride'], hparams['num_heads'], 
                            hparams['num_mha_layers'], hparams['d_ff'], hparams['length_seq'], 
                            hparams['alphabet_size'], dropout=hparams['drop'])
  
    self.encoder = PairedEncoder(hparams['d_embedding_cross'],  hparams['num_heads_cross'], hparams['d_ff_cross'],
                                      vh_encoder, vl_encoder, hparams['freeze_heavy'], hparams['freeze_light'],
                                      hparams['drop'], hparams['gate_bias'])

    # Paired decoder
    vh_decoder = Decoder(hparams['d_embedding'], hparams['kernel'], hparams['stride'], hparams['num_heads'], 
                            hparams['num_mha_layers'], hparams['d_ff'], hparams['length_seq'], 
                            hparams['alphabet_size'], dropout=hparams['drop'])
    
    vl_decoder = Decoder(hparams['d_embedding'], hparams['kernel'], hparams['stride'], hparams['num_heads'], 
                            hparams['num_mha_layers'], hparams['d_ff'], hparams['length_seq'], 
                            hparams['alphabet_size'], dropout=hparams['drop'])
  
    self.decoder = PairedDecoder(hparams['d_embedding_cross'],  hparams['num_heads_cross'], hparams['d_ff_cross'],
                                      vh_decoder, vl_decoder, hparams['freeze_heavy'], hparams['freeze_light'],
                                      hparams['drop'], hparams['gate_bias'])
    

    # Codebooks
    self.vh_vqvae = VectorQuantize(
            dim=hparams['d_embedding'],
            codebook_size=hparams['num_embeddings'],
            codebook_dim=hparams['embedding_dim_code_book'],
            decay=hparams['decay'],
            kmeans_init=True,
            commitment_weight=hparams['commitment_cost']
            )
    
    self.vl_vqvae = VectorQuantize(
            dim=hparams['d_embedding'],
            codebook_size=hparams['num_embeddings'],
            codebook_dim=hparams['embedding_dim_code_book'],
            decay=hparams['decay'],
            kmeans_init=True,
            commitment_weight=hparams['commitment_cost']
            )
    
    # Freezing
    if hparams['freeze_heavy']:  # Freeze VH model if requested
        for param in self.vh_vqvae.parameters():
            param.requires_grad = False
        for param in self.encoder.heavy_network.parameters():
            param.requires_grad = False
        for param in self.decoder.heavy_network.parameters():
            param.requires_grad = False

    if hparams['freeze_light']:  # Freeze VL model if requested
        for param in self.vl_vqvae.parameters():
            param.requires_grad = False
        for param in self.encoder.light_network.parameters():
            param.requires_grad = False
        for param in self.decoder.light_network.parameters():
            param.requires_grad = False

    # Pairing prediction
    self.linear_heavy_ppred = nn.Sequential(
       nn.Linear(hparams['d_embedding'], hparams['d_ppred_lr'], bias=False),
        nn.LayerNorm(hparams['d_ppred_lr'])
    )
    
    self.linear_light_ppred =  nn.Sequential(
       nn.Linear(hparams['d_embedding'], hparams['d_ppred_lr'], bias=False),
        nn.LayerNorm(hparams['d_ppred_lr'])
    )
    
    self.lr = nn.Linear(hparams['d_ppred_lr']*vh_encoder.l_red, 1)
    # self.lr3 = nn.Linear(hparams['d_ppred_lr']*vh_encoder.l_red, 3)

    # TRAINING
    self.learning_rate = hparams['learning_rate']
    self.bce_loss = nn.BCELoss()

    # LOSS 
    if "loss" in hparams: self.loss = hparams['loss']
    else: self.loss = 'mse'
    if "gamma" in hparams: self.gamma = hparams['gamma']
    else: self.gamma = 1
    if "lambda" in hparams: self.lambda_ = hparams['lambda']
    else: self.lambda_ = 1

    # REAL-TIME ANALYSIS
    self.validation_step_outputs = []
    self.fp_diverse_test = hparams['fp_diverse_test']
    self.fp_pssm = hparams['fp_pssm']

    self.fp_test = hparams['fp_test']

    if 'fp_test_1l_perh_matching' in hparams:
        self.fp_test_1l_perh_matching = hparams['fp_test_1l_perh_matching']
        self.fp_test_1l_perh_binned = hparams['fp_test_1l_perh_binned']

        self.fp_test_50l_perh_matching = hparams['fp_test_50l_perh_matching']
        self.fp_test_50l_perh_binned = hparams['fp_test_50l_perh_binned']

        self.fp_test_mismatch = hparams['fp_test_mismatch']

        self.fp_mouse = hparams['fp_mouse']
        self.fp_rat = hparams['fp_rat']

        self.imgt_human = hparams['fp_imgt_human']
        self.imgt_non_human = hparams['fp_imgt_non_human']

        self.ada = hparams['fp_ada']

    if 'lr_head' in hparams:
       self.lr_head = hparams['lr_head']
    else:
       self.lr_head = False


    self.batch_size = hparams['batch_size']
    self.zeta = hparams['zeta']

    if "conservation_index" in hparams:
      self.conservation_index = hparams["conservation_index"]
    else: 
      self.conservation_index = None

    if "caa_freq" in hparams:
      self.aa_freq = hparams["aa_freq"]
    else: 
      self.aa_freq = None

    if "weight_decay" in hparams:
       self.weight_decay = hparams["weight_decay"]
    else:
       self.weight_decay = 0

    self.save_hyperparameters(hparams)
    self._pretrained_loaded = False

  def on_train_start(self):
    if self.conservation_index:
      with open(self.conservation_index, "rb") as f:
          conservation_indices = pickle.load(f)
          self.conservation_indices = torch.tensor(1 - np.array(conservation_indices))

    if self.aa_freq:
      with open(self.aa_freq, "rb") as f:
          aa_freq = pickle.load(f)
          logging.info(f"Working with aa_freq order: {''.join(aa_freq.keys())}. Ensure this matches the dataloader's aa order.")
          self.aa_freq = torch.tensor(list(aa_freq.values()), dtype=torch.float32).T  # (input_seq_len, alphabet_size)
    
  def on_train_batch_start(self, batch, batch_idx):
        """Load pre-trained weights only at step 0"""

        self.batch_load_time = time.time() - self.load_start_time
        self.log(
            "train/load_time",
            self.batch_load_time,
            on_step=True,
            on_epoch=False,
            prog_bar=False,
        )
        self.step_start_time = time.time()

        if self.global_step == 0:
            print('LOADING THE PRE-TRAINED MODELS')
            print(f"Epoch {self.current_epoch} | global_step: {self.global_step} | batch_idx: {batch_idx}")
            self._pretrained_loaded = True

            vh_ckpt = torch.load(self.hparams["pre_trained_vh_ckpt"], map_location=self.device)
            vl_ckpt = torch.load(self.hparams["pre_trained_vl_ckpt"], map_location=self.device)

            self.encoder.heavy_network.load_state_dict({key.replace('encoder.',''): value for key, value in vh_ckpt["state_dict"].items() if key.startswith("encoder.")})
            self.encoder.light_network.load_state_dict({key.replace('encoder.',''): value for key, value in vl_ckpt["state_dict"].items() if key.startswith("encoder.")})
            
            self.decoder.heavy_network.load_state_dict({key.replace('decoder.',''): value for key, value in vh_ckpt["state_dict"].items() if key.startswith("decoder.")})
            self.decoder.light_network.load_state_dict({key.replace('decoder.',''): value for key, value in vl_ckpt["state_dict"].items() if key.startswith("decoder.")})

            self.vh_vqvae.load_state_dict({key.replace('vqvae.',''): value for key, value in vh_ckpt["state_dict"].items() if key.startswith("vqvae.")})
            self.vl_vqvae.load_state_dict({key.replace('vqvae.',''): value for key, value in vl_ckpt["state_dict"].items() if key.startswith("vqvae.")})

            if self.hparams['freeze_heavy']:  # Freeze VH model if requested
                for param in self.vh_vqvae.parameters():
                    param.requires_grad = False
                for param in self.encoder.heavy_network.parameters():
                    param.requires_grad = False
                for param in self.decoder.heavy_network.parameters():
                    param.requires_grad = False

            if self.hparams['freeze_light']:  # Freeze VL model if requested
                for param in self.vl_vqvae.parameters():
                    param.requires_grad = False
                for param in self.encoder.light_network.parameters():
                    param.requires_grad = False
                for param in self.decoder.light_network.parameters():
                    param.requires_grad = False

  def on_train_batch_end(self, outputs, batch, batch_idx: int):
        self.batch_step_time = time.time() - self.step_start_time
        self.log(
            "train/step_time",
            self.batch_step_time,
            on_step=True,
            on_epoch=False,
            prog_bar=False,
        )
        self.load_start_time = time.time()


  def forward(self, data) -> dict:

    vh_inputs, vh_m_inputs = data[0]
    vl_inputs, vl_m_inputs = data[1]

    B = vh_inputs.size(0)

    vh_x, vl_x = self.encoder(vh_m_inputs, vl_m_inputs)
    vh_vq_outputs = self.vh_vqvae(vh_x)
    vl_vq_outputs = self.vl_vqvae(vl_x)
    vh_x_recon, vl_x_recon, vh_x_pre_proj, vl_x_pre_proj = self.decoder(vh_vq_outputs['quantize_projected_out'], vl_vq_outputs['quantize_projected_out'])

    # Loss computing 
    vh_recon_error_pposi = self.calculate_recon_error_pposi(vh_inputs, vh_x_recon)
    vh_recon_error_pbe = torch.mean(vh_recon_error_pposi, dim=1)
    vh_lambda_vq_output = self.lambda_ * vh_vq_outputs['loss_vq_commit_pbe']
    vh_loss_pbe = torch.add(vh_recon_error_pbe, vh_lambda_vq_output)

    vl_recon_error_pposi = self.calculate_recon_error_pposi(vl_inputs, vl_x_recon)
    vl_recon_error_pbe = torch.mean(vl_recon_error_pposi, dim=1)
    vl_lambda_vq_output = self.lambda_ * vl_vq_outputs['loss_vq_commit_pbe']
    vl_loss_pbe = torch.add(vl_recon_error_pbe, vl_lambda_vq_output)

    inputs = torch.cat([vh_inputs,vl_inputs], dim=-2)
    x_recon = torch.cat([vh_x_recon,vl_x_recon], dim=-2)
    recon_error_pposi = torch.cat([vh_recon_error_pposi,vl_recon_error_pposi], dim=-1)
    loss_pbe = (vh_loss_pbe + vl_loss_pbe)/2
    recon_error_pbe = (vh_recon_error_pbe + vl_recon_error_pbe)/2

    mse_pposi = self.calculate_MSE_recon_error_pposi(inputs, x_recon)
    
    # Pairing prediction (positive)
    dot_paired = torch.flatten(
        self.linear_heavy_ppred(vh_x_pre_proj) * self.linear_light_ppred(vl_x_pre_proj),
        start_dim=1
    )
    paired_pred = torch.sigmoid(self.lr(dot_paired))
    target_paired = torch.ones((B, 1), device=paired_pred.device)

  
    if self.training:

        # Shuffling lights for contrastive learning 
        shuffled_indices = torch.randperm(B)
        shuffled_vl_m_inputs = vl_m_inputs[shuffled_indices]

        from_shuff_vh_x, shuffled_vl_x = self.encoder(vh_m_inputs, shuffled_vl_m_inputs)
        from_shuff_vh_vq_outputs = self.vh_vqvae(from_shuff_vh_x)
        shuffled_vl_vq_outputs = self.vl_vqvae(shuffled_vl_x)
        from_shuff_vh_x_recon, shuffled_vl_x_recon, from_shuff_vh_pre_proj, shuffled_vl_pre_proj = self.decoder(
            from_shuff_vh_vq_outputs['quantize_projected_out'],
            shuffled_vl_vq_outputs['quantize_projected_out']
        )

        # Shuffled negative prediction
        dot_NOT_paired = torch.flatten(
            self.linear_heavy_ppred(from_shuff_vh_pre_proj) * self.linear_light_ppred(shuffled_vl_pre_proj),
            start_dim=1
        )
        NOT_paired_pred = torch.sigmoid(self.lr(dot_NOT_paired))
        target_NOT_paired = torch.zeros((B, 1), device=paired_pred.device)

        # Humatch mismatch prediction
        vh_mismatch_inputs, vh_mismatch_m_inputs = data[2]
        vl_mismatch_inputs, vl_mismatch_m_inputs = data[3]
        
        from_mismatch_vh_x, mismatch_vl_x = self.encoder(vh_mismatch_m_inputs, vl_mismatch_m_inputs)
        from_mismatch_vh_vq_outputs = self.vh_vqvae(from_mismatch_vh_x)
        mismatch_vl_vq_outputs = self.vl_vqvae(mismatch_vl_x)
        from_mismatch_vh_x_recon, mismatch_vl_x_recon, from_mismatch_vh_pre_proj, mismatch_vl_pre_proj = self.decoder(
            from_mismatch_vh_vq_outputs['quantize_projected_out'],
            mismatch_vl_vq_outputs['quantize_projected_out']
        )

        dot_mismatch_paired = torch.flatten(
            self.linear_heavy_ppred(from_mismatch_vh_pre_proj) * self.linear_light_ppred(mismatch_vl_pre_proj),
            start_dim=1
        )
        mismatch_pred = torch.sigmoid(self.lr(dot_mismatch_paired))
        target_mismatch = torch.zeros((B, 1), device=paired_pred.device)

        # weighted BCE losses (if focal, use binary_focal_loss())
        pos_loss = self.bce_loss(paired_pred, target_paired)
        shuf_loss = self.bce_loss(NOT_paired_pred, target_NOT_paired)
        mismatch_loss = self.bce_loss(mismatch_pred, target_mismatch)

        ppred_loss = (
            self.hparams["weight_positive"] * pos_loss.mean()
            + self.hparams["weight_shuffled"] * shuf_loss.mean()
            + self.hparams["weight_mismatch"] * mismatch_loss.mean()
        ) / (self.hparams["weight_positive"] + self.hparams["weight_shuffled"] + self.hparams["weight_mismatch"])


        return {
            'vh_inputs': vh_inputs, # (batch_size, input_seq_len, alphabet_size)
            'vh_x_recon': vh_x_recon, # (batch_size, input_seq_len, alphabet_size)
            'vh_recon_error_pposi': vh_recon_error_pposi, # (batch_size, input_seq_len)
            'vh_recon_error_pbe': vh_recon_error_pbe, # (batch_size)
            'vh_loss_pbe': vh_loss_pbe, # (batch_size)
            **{f"vh_{key}": value for key, value in vh_vq_outputs.items()},

            'vl_inputs': vl_inputs, # (batch_size, input_seq_len, alphabet_size)
            'vl_x_recon': vl_x_recon, # (batch_size, input_seq_len, alphabet_size)
            'vl_recon_error_pposi': vl_recon_error_pposi, # (batch_size, input_seq_len)
            'vl_recon_error_pbe': vl_recon_error_pbe, # (batch_size)
            'vl_loss_pbe': vl_loss_pbe, # (batch_size)
            **{f"vl_{key}": value for key, value in vl_vq_outputs.items()},

            'inputs': inputs,
            'x_recon': x_recon, # (batch_size, input_seq_len, alphabet_size)
            'mse_pposi': mse_pposi, # (batch_size, input_seq_len)
            'recon_error_pposi': recon_error_pposi, # (batch_size, input_seq_len)
            'recon_error_pbe': recon_error_pbe,
            'loss_pbe': loss_pbe,

            'pairing_pred': paired_pred,
            'ppred_loss': ppred_loss,

            'pos_loss': pos_loss.mean(),
            'shuf_loss': shuf_loss.mean(),
            'mismatch_loss': mismatch_loss.mean(),
        }
        
    else: # Only for deployable version
       
       ppred_loss = torch.Tensor([0])

       return {
            'vh_inputs': vh_inputs, # (batch_size, input_seq_len, alphabet_size)
            'vh_x_recon': vh_x_recon, # (batch_size, input_seq_len, alphabet_size)
            'vh_recon_error_pposi': vh_recon_error_pposi, # (batch_size, input_seq_len)
            'vh_recon_error_pbe': vh_recon_error_pbe, # (batch_size)
            'vh_loss_pbe': vh_loss_pbe, # (batch_size)
            **{f"vh_{key}": value for key, value in vh_vq_outputs.items()},

            'vl_inputs': vl_inputs, # (batch_size, input_seq_len, alphabet_size)
            'vl_x_recon': vl_x_recon, # (batch_size, input_seq_len, alphabet_size)
            'vl_recon_error_pposi': vl_recon_error_pposi, # (batch_size, input_seq_len)
            'vl_recon_error_pbe': vl_recon_error_pbe, # (batch_size)
            'vl_loss_pbe': vl_loss_pbe, # (batch_size)
            **{f"vl_{key}": value for key, value in vl_vq_outputs.items()},

            'inputs': inputs,
            'x_recon': x_recon, # (batch_size, input_seq_len, alphabet_size)
            'mse_pposi': mse_pposi, # (batch_size, input_seq_len)
            'recon_error_pposi': recon_error_pposi, # (batch_size, input_seq_len)
            'recon_error_pbe': recon_error_pbe,
            'loss_pbe': loss_pbe,

            'pairing_pred': paired_pred,
            'ppred_loss': ppred_loss,

            'vh_x_encoder_pre_vq': vh_x,
            'vl_x_encoder_pre_vq': vl_x,
            'vh_quantize_projected_out': vh_vq_outputs['quantize_projected_out'],
            'vl_quantize_projected_out': vl_vq_outputs['quantize_projected_out'],
            'vh_x_decoder_pre_proj': vh_x_pre_proj,
            'vl_x_decoder_pre_proj': vl_x_pre_proj,
       }
       
       

  def calculate_MSE_recon_error_pposi(self, inputs, x_recon, conservation_index=False, focal=False, alpha=False):
    recon_error_pres_pposi = F.mse_loss(x_recon, inputs, reduction='none')
    recon_error_pposi = torch.mean(recon_error_pres_pposi, dim=-1)
    
    if conservation_index:
        self.conservation_indices = self.conservation_indices.to(recon_error_pposi.device)
        conservation_term = self.conservation_indices ** self.gamma
        recon_error_pposi *= conservation_term
    
    if focal: 
        p_true = torch.gather(x_recon, -1, inputs.argmax(dim=-1, keepdim=True)).squeeze(-1)
        focal_term = (1 - p_true) ** self.gamma
        recon_error_pposi *= focal_term

    if alpha:
        self.aa_freq = self.aa_freq.to(recon_error_pposi.device)
        inputs = inputs.to(recon_error_pposi.device)
        index_true = inputs.argmax(dim=-1)
        true_aa_freq = self.aa_freq[torch.arange(recon_error_pposi.shape[-1]), index_true]
        alpha_term = (1 - true_aa_freq)
        recon_error_pposi *= alpha_term
    
    return recon_error_pposi

  def calculate_CE_recon_error_pposi(self, inputs, x_recon, conservation_index=False, focal=False):
    x_recon = x_recon.reshape(x_recon.shape[0], x_recon.shape[2], x_recon.shape[1])
    inputs = inputs.reshape(inputs.shape[0], inputs.shape[2], inputs.shape[1])
    recon_error_pposi = F.cross_entropy(x_recon, inputs, reduction='none')

    if conservation_index:
        self.conservation_indices = self.conservation_indices.to(recon_error_pposi.device)
        conservation_term = self.conservation_indices ** self.gamma
        recon_error_pposi *= conservation_term
    
    if focal: 
        p_true = torch.gather(x_recon, -1, inputs.argmax(dim=-1, keepdim=True)).squeeze(-1)
        focal_term = (1 - p_true) ** self.gamma
        recon_error_pposi *= focal_term

    return recon_error_pposi

  def calculate_recon_error_pposi(self, inputs, x_recon):
    # MSE
    if self.loss == 'mse':
        recon_error_pposi = self.calculate_MSE_recon_error_pposi(inputs, x_recon)
    elif self.loss == 'conservation_mse':
        recon_error_pposi = self.calculate_MSE_recon_error_pposi(inputs, x_recon, conservation_index=True)
    elif self.loss == 'focal_mse':
        recon_error_pposi = self.calculate_MSE_recon_error_pposi(inputs, x_recon, focal=True)
    elif self.loss == 'alpha_focal_mse':
        recon_error_pposi = self.calculate_MSE_recon_error_pposi(inputs, x_recon, focal=True, alpha=True)

    # CE
    elif self.loss == 'ce':
        recon_error_pposi = self.calculate_CE_recon_error_pposi(inputs, x_recon)
    elif self.loss == 'conservation_ce':
        recon_error_pposi = self.calculate_CE_recon_error_pposi(inputs, x_recon, conservation_index=True)
    elif self.loss == 'focal_ce':
        recon_error_pposi = self.calculate_CE_recon_error_pposi(inputs, x_recon, focal=True)

    else:
        raise ValueError(f"Unsupported loss type: {self.loss}")
    
    return recon_error_pposi

  def configure_optimizers(self):

    if self.lr_head is False:
        optim_groups = list(self.encoder.parameters()) + \
                        list(self.decoder.parameters()) + \
                        list(self.vh_vqvae.parameters()) + \
                        list(self.vl_vqvae.parameters()) + \
                        \
                        list(self.linear_heavy_ppred.parameters()) + \
                        list(self.linear_light_ppred.parameters()) + \
                        list(self.lr.parameters())

        return torch.optim.AdamW(optim_groups, lr=self.learning_rate, weight_decay=self.weight_decay)
    
    else:
        head = list(self.linear_heavy_ppred.parameters()) + \
            list(self.linear_light_ppred.parameters()) + \
            list(self.lr.parameters())

        backbone = list(self.encoder.parameters()) + \
                list(self.decoder.parameters()) + \
                list(self.vh_vqvae.parameters()) + \
                list(self.vl_vqvae.parameters())

        return torch.optim.AdamW(
            [
                {"params": backbone, "lr": self.learning_rate, "weight_decay": self.weight_decay},
                {"params": head,     "lr": self.lr_head, "weight_decay": self.weight_decay},
            ]
        )

  def training_step(self, batch, batch_idx) -> float:
    if batch_idx % 1000 == 0:
      current_time = datetime.now().strftime("%H:%M:%S %d-%m-%Y")
      logging.info(f"Training on batch {batch_idx} at {current_time}")
    
    vqvae_output = self(batch)

    loss_vqvae = torch.mean(vqvae_output['loss_pbe'])
    self.log("train_loss_vqvae", loss_vqvae, on_step=True, on_epoch=True, prog_bar=True, logger=True)

    loss_vqvae_vh = torch.mean(vqvae_output['vh_loss_pbe'])
    self.log("train_loss_vqvae_vh", loss_vqvae_vh, on_step=True, on_epoch=True, prog_bar=True, logger=True)
    loss_vqvae_vl = torch.mean(vqvae_output['vl_loss_pbe'])
    self.log("train_loss_vqvae_vl", loss_vqvae_vl, on_step=True, on_epoch=True, prog_bar=True, logger=True)

    loss_vq_commit_vh = torch.mean(vqvae_output['vh_loss_vq_commit_pbe'])
    self.log("train_loss_vq_commit_vh", loss_vq_commit_vh, on_step=True,on_epoch=True,prog_bar=True, logger=True)
    loss_vq_commit_vl = torch.mean(vqvae_output['vl_loss_vq_commit_pbe'])
    self.log("train_loss_vq_commit_vl", loss_vq_commit_vl, on_step=True,on_epoch=True,prog_bar=True, logger=True)

    nmse_accuracy_vh = torch.mean(vqvae_output['vh_recon_error_pbe'])
    self.log("train_loss_nmse_recons_vh", nmse_accuracy_vh, on_step=True, on_epoch=True,prog_bar=True, logger=True)
    nmse_accuracy_vl = torch.mean(vqvae_output['vl_recon_error_pbe'])
    self.log("train_loss_nmse_recons_vl", nmse_accuracy_vl, on_step=True, on_epoch=True,prog_bar=True, logger=True)

    vh_perplexity = vqvae_output['vh_perplexity']
    self.log("train_perplexity_vh", vh_perplexity, on_step=True, on_epoch=True,prog_bar=True, logger=True)
    vl_perplexity = vqvae_output['vl_perplexity']
    self.log("train_perplexity_vl", vl_perplexity, on_step=True, on_epoch=True,prog_bar=True, logger=True)

    ppred_loss = vqvae_output['ppred_loss']
    self.log("ppred_loss", ppred_loss, on_step=True, on_epoch=True,prog_bar=True, logger=True)

    total_loss = loss_vqvae + self.zeta * ppred_loss
    self.log("train_loss_vqvae_pairing", total_loss, on_step=True, on_epoch=True,prog_bar=True, logger=True)

    self.log("pos_loss", vqvae_output['pos_loss'], on_step=True, on_epoch=True,prog_bar=True, logger=True)
    self.log("shuf_loss", vqvae_output['shuf_loss'], on_step=True, on_epoch=True,prog_bar=True, logger=True)
    self.log("mismatch_loss", vqvae_output['mismatch_loss'], on_step=True, on_epoch=True,prog_bar=True, logger=True)

    self.log('pairing_pred', vqvae_output['pairing_pred'].mean(), on_step=True, on_epoch=True,prog_bar=True, logger=True)

    return total_loss

  def on_train_epoch_start(self) -> None:
    current_time = datetime.now().strftime("%H:%M:%S %d-%m-%Y")
    logging.info(f"Training Epoch {self.current_epoch} started at {current_time}")
    self.load_start_time = time.time()
    return

  def on_train_epoch_end(self) -> None:
    current_time = datetime.now().strftime("%H:%M:%S %d-%m-%Y")
    logging.info(f"Training Epoch {self.current_epoch} ended at {current_time}")
    return 

  def validation_step(self, batch, batch_idx) -> None:
    """
    Perform a single validation step for a given batch during the validation epoch.

    This method evaluates a batch of sequences using the AbNatiV model, computes various nativeness scores,
    and appends the results to the validation step outputs.

    Notes
    -----
    The method performs the following steps:
    1. Initializes a dictionary to store the mean scores for the batch.
    2. Calls `get_abnativ_scores_per_batch` to evaluate the batch and compute nativeness scores.
    3. Constructs an output dictionary containing the current epoch, batch index, model mean outputs, and VHH scores.
    4. Appends the output dictionary to `self.validation_step_outputs`.
    """
    
    if batch_idx % 1000 == 0:
      current_time = datetime.now().strftime("%H:%M:%S %d-%m-%Y")
      logging.info(f"Validation on batch {batch_idx} at {current_time}")

    scored_data_dict_mean = defaultdict(list)
    # Get abnativ scores for this batch
    model_output = get_abnativ_scores_per_batch(
      model=self, 
      model_type='None', 
      batch=batch,
      scored_data_dict_mean=scored_data_dict_mean
    )

    mean_loss = torch.mean(model_output['loss_pbe'])
    mean_recon_error = torch.mean(model_output["recon_error_pbe"])

    inputs_seq = torch.argmax(model_output["inputs"], dim=-1)
    x_recon_seq = torch.argmax(model_output["x_recon"], dim=-1)
    mean_accuracy = torch.mean((inputs_seq == x_recon_seq).float()).item()

    mean_ppred_loss = torch.mean(model_output['ppred_loss'])
    mean_pairing_pred = torch.mean(model_output['pairing_pred'])

    mean_total_loss = mean_loss + self.zeta * mean_ppred_loss


    mean_output = {
      "mean_total_loss": mean_total_loss,
      "mean_loss": mean_loss,
      "mean_recon_error": mean_recon_error,
      "mean_accuracy": mean_accuracy,
      "mean_ppred_loss": mean_ppred_loss,
      "mean_pairing_pred": mean_pairing_pred
    }

    score_to_keep = "AbNatiV None Score"
    scored_data_dict_mean = defaultdict(list, {score_to_keep: scored_data_dict_mean[score_to_keep]})

    output = {
      "epoch": self.current_epoch,
      "batch_num": batch_idx,
      "model_output": mean_output,
      "scores": scored_data_dict_mean,
    }
    self.validation_step_outputs.append(output)
    return 

  def log_and_save_validation_metrics(self, output_dir: str) -> None:
    # Extract outputs for the current epoch. Although validation_step_outputs should only contain scores for the current epoch, 
    # we filter them to ensure consistency and robustness.
    model_outputs = [
      out["model_output"]
      for out in self.validation_step_outputs
      if out["epoch"] == self.current_epoch
    ]

    val_mean_total_loss = torch.Tensor([out["mean_total_loss"] for out in model_outputs])
    total_val_mean_total_loss = torch.mean(val_mean_total_loss)
    self.log('val_total_loss', total_val_mean_total_loss, on_epoch=True, logger=True)

    val_losses = torch.Tensor([out["mean_loss"] for out in model_outputs])
    total_val_loss = torch.mean(val_losses)
    self.log('val_vqvae_loss', total_val_loss, on_epoch=True, logger=True)

    val_accuracies = torch.Tensor([out['mean_recon_error'] for out in model_outputs])
    total_val_accuracy = torch.mean(val_accuracies)
    self.log('val_nmse_accuracy', total_val_accuracy, on_epoch=True, logger=True)

    accuracies = torch.Tensor([out['mean_accuracy'] for out in model_outputs])
    total_val_accuracy = torch.mean(accuracies)
    self.log('val_accuracy', total_val_accuracy, on_epoch=True, logger=True)

    val_ppred_loss = torch.Tensor([out['mean_ppred_loss'] for out in model_outputs])
    total_val_ppred_loss = torch.mean(val_ppred_loss)
    self.log('val_ppred_loss', total_val_ppred_loss, on_epoch=True, logger=True)

    val_mean_pairing_pred = torch.Tensor([out['mean_pairing_pred'] for out in model_outputs])
    total_mean_pairing_pred = torch.mean(val_mean_pairing_pred)
    self.log('val_mean_pairing_pred', total_mean_pairing_pred, on_epoch=True, logger=True)

    # Extract the specified keys from each dictionary in model_outputs
    with open(f'{output_dir}/val_model_outputs.pkl', 'wb') as f:
        pickle.dump(model_outputs, f)

  def save_scores_and_process_pssm(self, output_dir: str) -> None:
    """
    Analyze and save validation scores, score PSSM using the current model, and generate ROC and PR plots.

    This function performs the following actions:
    1. Analyzes and saves validation scores stored in self.validation_step_outputs
    2. Scores PSSM (Position-Specific Scoring Matrix) using the current model and saves the results.
    3. Runs ROC (Receiver Operating Characteristic) and PR (Precision-Recall) analyses between validation scores and PSSM scores, and saves the resulting plots.

    Parameters
    ----------
    output_dir : str
        The path to the output directory where results and plots will be saved.
    """

    # Extract scores for the current epoch. Although validation_step_outputs should only contain scores for the current epoch, 
    # we filter them to ensure consistency and robustness.
    epoch_scores = [
        out["scores"]
        for out in self.validation_step_outputs
        if out["epoch"] == self.current_epoch
    ]

    # Combine scores from all batches into a single dictionary for the current epoch.
    scored_data_dict_mean = defaultdict(list)
    for scores in epoch_scores:
        for key in scores:
            scored_data_dict_mean[key].extend(scores[key])

    # Create a pandas DataFrame for the epoch scores and save it as 'val_scores.csv' in the specified directory.
    df_mean_epoch = pd.DataFrame.from_dict(scored_data_dict_mean)
    df_mean_epoch.to_csv(f"{output_dir}/val_scores.csv")

    ### VALIDATION vs others ###
    pr_auc, roc_auc = calculate_pssm_aucs_paired_with_mismatch(
      model=self, 
      model_type='None', 
      df_mean_val=df_mean_epoch, 
      fp_pssm=self.fp_pssm, 
      batch_size=self.batch_size, 
      output_dir=output_dir
    )

    self.log('PR Val-PSSM', torch.tensor(pr_auc), on_epoch=True, logger=True)

    pr_auc, roc_auc = calculate_pssm_aucs_paired_with_mismatch(
      model=self, 
      model_type='None', 
      df_mean_val=df_mean_epoch, 
      fp_pssm=self.fp_mouse, 
      batch_size=self.batch_size, 
      output_dir=output_dir
    )

    self.log('PR Val-Mouse', torch.tensor(pr_auc), on_epoch=True, logger=True)

    pr_auc, roc_auc = calculate_pssm_aucs_paired_with_mismatch(
      model=self, 
      model_type='None', 
      df_mean_val=df_mean_epoch, 
      fp_pssm=self.fp_rat, 
      batch_size=self.batch_size, 
      output_dir=output_dir
    )

    self.log('PR Val-Rat', torch.tensor(pr_auc), on_epoch=True, logger=True)

    pr_auc, roc_auc = calculate_pssm_aucs_paired_with_mismatch(
      model=self, 
      model_type='None', 
      df_mean_val=df_mean_epoch, 
      fp_pssm=self.fp_diverse_test, 
      batch_size=self.batch_size, 
      output_dir=output_dir
    )

    self.log('PR Val-Diverse', torch.tensor(pr_auc), on_epoch=True, logger=True)


    pr_auc, roc_auc = calculate_pssm_aucs_paired_with_mismatch(
      model=self, 
      model_type='None', 
      df_mean_val=df_mean_epoch, 
      fp_pssm=self.fp_test_mismatch, 
      batch_size=self.batch_size, 
      output_dir=output_dir
    )

    self.log('PR Val-Mismatch', torch.tensor(pr_auc), on_epoch=True, logger=True)


    # IMGT TEST
    df_mean_imgt_human = evaluate_paired_pssm_with_mismatch(self, 'None', self.imgt_human, self.batch_size)

    pr_auc, roc_auc = calculate_pssm_aucs_paired_with_mismatch(
      model=self, 
      model_type='None', 
      df_mean_val=df_mean_imgt_human, 
      fp_pssm=self.imgt_non_human, 
      batch_size=self.batch_size, 
      output_dir=output_dir
    )

    self.log('PR IMGT Human-Non-Human', torch.tensor(pr_auc), on_epoch=True, logger=True)

    # ADA TEST
    df_mean_ada = evaluate_paired_pssm_with_mismatch(self, 'None', self.ada, self.batch_size)
    ada_scores = df_mean_ada["AbNatiV None Score"]
    df_ada_immuno = pd.read_csv(self.ada)
    immunogenicities = df_ada_immuno['Immunogenicity']

    correlation, p_value = pearsonr(ada_scores, immunogenicities)

    self.log('ADA pearson correlation', torch.tensor(correlation), on_epoch=True, logger=True)


    ### DIVERSE vs others ###
    df_mean_diverse_test = evaluate_paired_pssm_with_mismatch(self, 'None', self.fp_diverse_test, self.batch_size)


    pr_auc, roc_auc = calculate_pssm_aucs_paired_with_mismatch(
      model=self, 
      model_type='None', 
      df_mean_val=df_mean_diverse_test, 
      fp_pssm=self.fp_pssm, 
      batch_size=self.batch_size, 
      output_dir=output_dir
    )

    self.log('PR Diverse-PSSM', torch.tensor(pr_auc), on_epoch=True, logger=True)

    pr_auc, roc_auc = calculate_pssm_aucs_paired_with_mismatch(
      model=self, 
      model_type='None', 
      df_mean_val=df_mean_diverse_test, 
      fp_pssm=self.fp_mouse, 
      batch_size=self.batch_size, 
      output_dir=output_dir
    )

    self.log('PR Diverse-Mouse', torch.tensor(pr_auc), on_epoch=True, logger=True)

    pr_auc, roc_auc = calculate_pssm_aucs_paired_with_mismatch(
      model=self, 
      model_type='None', 
      df_mean_val=df_mean_diverse_test, 
      fp_pssm=self.fp_rat, 
      batch_size=self.batch_size, 
      output_dir=output_dir
    )

    self.log('PR Diverse-Rat', torch.tensor(pr_auc), on_epoch=True, logger=True)

    


    ## PAIRING PERFORMANCES OF THE PAIRING VQ-VAE RECONSRUCTION ##
    # Pairing info 
    scores_p_heavy_true = evaluate_paired_pssm_with_mismatch(self, 'None', self.fp_test_1l_perh_matching, self.batch_size, aho_posi_whole=range(1,150))["AbNatiV None Score"]
    scores_p_heavy_random = evaluate_paired_pssm_with_mismatch(self, 'None', self.fp_test_1l_perh_binned, self.batch_size, aho_posi_whole=range(1,150))["AbNatiV None Score"]
    c_correct_pairing = np.sum(np.array(scores_p_heavy_true)[:len(scores_p_heavy_random)] > np.array(scores_p_heavy_random))/len(scores_p_heavy_random)

    self.log('[VQ-VAE Reconstruction] Correct pairing prediction Test 1:1 (on 10000H) [Vgene BINNED]', torch.tensor(c_correct_pairing), on_epoch=True, logger=True)

    mean_diff = np.mean(np.abs(np.array(scores_p_heavy_true)[:len(scores_p_heavy_random)] - np.array(scores_p_heavy_random)))
    self.log('[VQ-VAE Reconstruction] Mean Abs Diff Paired-Unpaired', torch.tensor(mean_diff), on_epoch=True, logger=True)

    # Pairing info per 50 (binned)
    scores_p_heavy_random_50_matching = evaluate_paired_pssm_with_mismatch(self, 'None', self.fp_test_50l_perh_matching, self.batch_size, aho_posi_whole=range(1,150))["AbNatiV None Score"]
    scores_p_heavy_random_50 = evaluate_paired_pssm_with_mismatch(self, 'None', self.fp_test_50l_perh_binned, self.batch_size, aho_posi_whole=range(1,150))["AbNatiV None Score"]
    max_heavy_seqs = 1000

    scores_p_heavy_random_50_mean = list()
    for i in range(0, len(scores_p_heavy_random_50), 50):
        scores_p_heavy_random_50_mean.append(np.mean(scores_p_heavy_random_50[i:i+50]))

    c_correct_pairing = np.sum(np.array(scores_p_heavy_random_50_matching[:max_heavy_seqs]) > np.array(scores_p_heavy_random_50_mean))/len(scores_p_heavy_random_50_mean)
    self.log('[VQ-VAE Reconstruction] Correct pairing prediction Test 50L per H (on 1000H) [Vgene BINNED]', torch.tensor(c_correct_pairing), on_epoch=True, logger=True)

    ## PAIRED vs MISMATCH ##
    df_mean_paired_test = evaluate_paired_pssm_with_mismatch(self, 'None', self.fp_test, self.batch_size)
    df_mean_mismatch_test = evaluate_paired_pssm_with_mismatch(self, 'None', self.fp_test_mismatch, self.batch_size)
    paired_scores = df_mean_paired_test["AbNatiV None Score"]
    mismatch_scores = df_mean_mismatch_test["AbNatiV None Score"]
    mean_diff = np.mean(paired_scores) - np.mean(mismatch_scores)
    self.log("Score separation Paired-Mismatch (mean)", torch.tensor(mean_diff), on_epoch=True, logger=True)

    pr_auc, roc_auc = calculate_pssm_aucs_paired_with_mismatch(
        model=self,
        model_type='None',
        df_mean_val=df_mean_paired_test,
        fp_pssm=self.fp_test_mismatch,
        batch_size=self.batch_size,
        output_dir=output_dir
    )
    self.log("PR AUC Paired vs Mismatch", torch.tensor(pr_auc), on_epoch=True, logger=True)
    self.log("ROC AUC Paired vs Mismatch", torch.tensor(roc_auc), on_epoch=True, logger=True)


    ## PAIRING PERFORMANCES OF THE PAIRING PREDICTION HEAD ##
    # Pairing prediction head per 1 (binned)
    paired_pred_test_matching = evaluate_pairing_pred_with_mismatch(self, 'None', self.fp_test_1l_perh_matching, self.batch_size)
    paired_pred_shuffled = evaluate_pairing_pred_with_mismatch(self, 'None', self.fp_test_1l_perh_binned, self.batch_size)

    c_correct_pairing = np.sum(np.array(paired_pred_test_matching)[:len(paired_pred_shuffled)] > np.array(paired_pred_shuffled))/len(paired_pred_test_matching)
    self.log('[Prediction head] Correct pairing prediction Test 1:1 (on 10000H) [Vgene BINNED]', torch.tensor(c_correct_pairing), on_epoch=True, logger=True)

    # Pairing prediction head per 50 (binned)
    paired_pred_test_50matching = evaluate_pairing_pred_with_mismatch(self, 'None', self.fp_test_50l_perh_matching, self.batch_size)
    paired_pred_50_shuffled_binned = evaluate_pairing_pred_with_mismatch(self, 'None', self.fp_test_50l_perh_binned, self.batch_size)

    scores_paring_pred_heavy_random_50_mean_perbine = list()
    for i in range(0, len(paired_pred_50_shuffled_binned), 50):
        scores_paring_pred_heavy_random_50_mean_perbine.append(np.mean(paired_pred_50_shuffled_binned[i:i+50]))

    c_correct_pairing = np.sum(np.squeeze(np.array(paired_pred_test_50matching[:max_heavy_seqs])) > np.array(scores_paring_pred_heavy_random_50_mean_perbine))/len(scores_paring_pred_heavy_random_50_mean_perbine)
    self.log('[Prediction head] Correct pairing prediction Test 50L per H (on 1000H) [Vgene BINNED]', torch.tensor(c_correct_pairing), on_epoch=True, logger=True)

    # Compute confusion matrix
    label_pred_pairing_test_matching = list(np.array(paired_pred_test_matching) > 0.5)
    label_pred_pairing_test_shuffle = list(np.array(paired_pred_shuffled) > 0.5)

    tp, fn, fp, tn = confusion_matrix([True]*len(label_pred_pairing_test_matching)+[False]*len(label_pred_pairing_test_shuffle), 
                            label_pred_pairing_test_matching+label_pred_pairing_test_shuffle,
                            labels=[True,False]).ravel()
    
    self.log('[Prediction head] FN rate Test 1:1 (on 10000H) [Vgene BINNED]', torch.tensor(fn/len(label_pred_pairing_test_matching)), on_epoch=True, logger=True)
    self.log('[Prediction head] FP rate Test 1:1 (on 10000H) [Vgene BINNED]', torch.tensor(fp/len(label_pred_pairing_test_shuffle)), on_epoch=True, logger=True)



    ## ADD MISMATCH ##
    label_paired_pred_mismatch = evaluate_pairing_pred_with_mismatch(self, 'None', self.fp_test_mismatch, self.batch_size)
    nb_bellow_05_mismatch = sum(np.array(label_paired_pred_mismatch) < 0.5)

    self.log('[Prediction head] Ratio mismatch < 0.5 (on 50000H)', torch.tensor(nb_bellow_05_mismatch/len(label_paired_pred_mismatch)), on_epoch=True, logger=True)

    # Compute AUC Mismatch - Test 
    label_paired_pred_all_test = evaluate_pairing_pred_with_mismatch(self, 'None', self.fp_test, self.batch_size)
    
    combined_scores = pd.concat([pd.DataFrame(label_paired_pred_all_test), pd.DataFrame(label_paired_pred_mismatch)])
    true_labels = pd.Series([1] * len(label_paired_pred_all_test) + [0] * len(label_paired_pred_mismatch))
    precision, recall, thresholds = precision_recall_curve(
        true_labels, combined_scores
    )
    pr_auc = auc(recall, precision)

    self.log('[Prediction head] PR-AUC Test vs mismatch (on 50000H)', torch.tensor(pr_auc), on_epoch=True, logger=True)


    ## ADD PSSM ##
    label_paired_pred_pssm = evaluate_pairing_pred_with_mismatch(self, 'None', self.fp_pssm, self.batch_size)
    nb_bellow_05_pssm= sum(np.array(label_paired_pred_pssm) < 0.5)

    self.log('[Prediction head] Ratio PSSM < 0.5 (on 10000H)', torch.tensor(nb_bellow_05_pssm/len(label_paired_pred_pssm)), on_epoch=True, logger=True)

    ## ADD Mice ##
    label_paired_pred_mouse = evaluate_pairing_pred_with_mismatch(self, 'None', self.fp_mouse, self.batch_size)
    nb_bellow_05_mouse = sum(np.array(label_paired_pred_mouse) < 0.5)

    self.log('[Prediction head] Ratio Mouse < 0.5 (on 10000H)', torch.tensor(nb_bellow_05_mouse/len(label_paired_pred_mouse)), on_epoch=True, logger=True)

    ## ADD Rat ##
    label_paired_pred_rat = evaluate_pairing_pred_with_mismatch(self, 'None', self.fp_rat, self.batch_size)
    nb_bellow_05_rat = sum(np.array(label_paired_pred_rat) < 0.5)

    self.log('[Prediction head] Ratio Rat < 0.5 (on 10000H)', torch.tensor(nb_bellow_05_rat/len(label_paired_pred_rat)), on_epoch=True, logger=True)


  def on_validation_epoch_end(self) -> None:
    """
    Handles the end of a validation epoch.

    This method performs the following actions at the end of each validation epoch:
    1. Logs the current time and epoch number.
    2. Creates an output directory for the current epoch if it doesn't already exist.
    3. Logs and saves validation metrics to the output directory.
    4. Analyzes and saves validation scores stored in self.validation_step_outputs in a .pkl file.
       The saved data includes:
       {
           "accuracy": <accuracy_value>,
           "filtered_model_outputs": [
               {'loss_pbe': <value>, 'recon_error_pbe': <value>, 'perplexity': <value>},
               ...
           ]
       }
    5. Scores PSSM (Position-Specific Scoring Matrix) using the current model and saves the results.
    6. Runs ROC (Receiver Operating Characteristic) and PR (Precision-Recall) analyses between validation scores and PSSM scores, and saves the resulting plots.
    7. Clears the validation step outputs to prepare for the next epoch.
    """
    current_time = datetime.now().strftime("%H:%M:%S %d-%m-%Y")
    logging.info(f"Validation Epoch {self.current_epoch} finished at {current_time}")

    output_dir = f"analysis/{self.run_name}/{self.current_epoch}"
    if not os.path.exists(output_dir):
        os.makedirs(output_dir, exist_ok=True)
  
    self.log_and_save_validation_metrics(output_dir)
    self.save_scores_and_process_pssm(output_dir)

    self.validation_step_outputs = []

    return

# fmt: on



"""
 Copyright 2023. Aubin Ramon, Oliver Wissett and Pietro Sormanni. CC BY-NC-SA 4.0
"""

import argparse
import os
import subprocess
import tarfile
from typing import List


ZENODO_RECORD = "17295347"  # ← your actual Zenodo record ID
BASE_URL = f"https://zenodo.org/record/{ZENODO_RECORD}/files"

ANTIBODYBUILDER3_URL = "https://zenodo.org/records/11354577/files/output.tar.gz"
ANTIBODYBUILDER3_CKPT = "plddt-loss/best_second_stage.ckpt"
ANTIBODYBUILDER3_TARGET = "antibodybuilder3.ckpt"

# Check what OS we are on
def get_platform():
    return os.uname().sysname

if get_platform() == "Linux" or get_platform() == "Darwin":
    # NOTE: Maybe use /usr/local/share/abnativ/models instead?
    PRETRAINED_MODELS_DIR = os.path.expanduser("~/.abnativ/models/pretrained_models")
elif get_platform() == "Windows":
    PRETRAINED_MODELS_DIR = os.path.expanduser(
        "~\\AppData\\Local\\abnativ\\models\\pretrained_models"
    )
else:
    raise Exception("Unsupported OS")



def ensure_zenodo_models(filenames: List[str], do_force_update : bool, 
                         model_dir: str = PRETRAINED_MODELS_DIR):
    '''If do_force_update will force redowloading every file.'''

    os.makedirs(model_dir, exist_ok=True)

    if do_force_update: 
        missing = filenames
        print(f"Forcing update: downloading all the {len(missing)} files from Zenodo…")
    
    else:
        missing = []
        for fname in filenames:
            local_path = os.path.join(model_dir, fname)
            if not os.path.isfile(local_path):
                missing.append(fname)
            else:
                print(f"✔ {fname} already present")

        if not missing:
            print("All model files are present; nothing to download.")
            return

        print(f"Downloading {len(missing)} missing files from Zenodo…")

    for fname in missing:
        url = f"{BASE_URL}/{fname}?download=1"
        target = os.path.join(model_dir, fname)
        cmd = [
            "wget",
            "--content-disposition",
            "--no-cache", 
            "-O", target,
            url
        ]
        print(" ".join(cmd))
        subprocess.check_call(cmd)
        print(f"✔ Downloaded {fname}")

def ensure_antibodybuilder3(do_force_update: bool,
                             model_dir: str = PRETRAINED_MODELS_DIR):
    """Download output.tar.gz from Zenodo, extract best_second_stage.ckpt,
    rename it to antibodybuilder3.ckpt, and clean up."""

    target = os.path.join(model_dir, ANTIBODYBUILDER3_TARGET)

    if not do_force_update and os.path.isfile(target):
        print(f"✔ {ANTIBODYBUILDER3_TARGET} already present")
        return

    os.makedirs(model_dir, exist_ok=True)

    tarball = os.path.join(model_dir, "output.tar.gz")
    print(f"Downloading {ANTIBODYBUILDER3_URL} …")
    cmd = ["wget", "--no-cache", "-O", tarball, ANTIBODYBUILDER3_URL]
    print(" ".join(cmd))
    subprocess.check_call(cmd)

    print(f"Extracting {ANTIBODYBUILDER3_CKPT} …")
    with tarfile.open(tarball, "r:gz") as tar:
        member = tar.getmember(ANTIBODYBUILDER3_CKPT)
        # Extract just this one file flat into model_dir
        member.name = os.path.basename(member.name)  # strip subdirs
        tar.extract(member, path=model_dir)

    # Rename to antibodybuilder3.ckpt
    extracted = os.path.join(model_dir, os.path.basename(ANTIBODYBUILDER3_CKPT))
    os.rename(extracted, target)
    print(f"✔ Saved as {ANTIBODYBUILDER3_TARGET}")

    # Clean up tarball
    os.remove(tarball)
    print("✔ Cleaned up output.tar.gz")

def init(args: argparse.Namespace):
    expected = [
        "vhh2_model.ckpt",
        "vh2_model.ckpt",
        "vl2_model.ckpt",
        "vpaired2_model.ckpt",
        "vlambda_model.ckpt",
        "vkappa_model.ckpt",
        "vh_model.ckpt",
        "vhh_model.ckpt",
        "vh2_rhesus_model.ckpt",
    ]
    ensure_zenodo_models(expected, do_force_update=args.force_update)
    ensure_antibodybuilder3(do_force_update=args.force_update)
    print(f"✅ PRETRAINED_MODELS_DIR set to: {PRETRAINED_MODELS_DIR}")


# (c) 2023 Sormannilab and Aubin Ramon
#
# Humanisation of VHH sequences using AbNatiV VH and VHH assesments.
#
# ============================================================================

from .model.utils import is_protein
from .model.onehotencoder import alphabet
from .model.scoring_functions import fr_aho_indices
from .humanisation.vhh_humanisation_functions import abnativ_vhh_humanisation

import os
import argparse

from Bio import SeqIO

def run(args: argparse.Namespace):

    fp_fa_or_seq = args.input_filepath_or_seq

    # If input is a single sequence
    if not os.path.isfile(fp_fa_or_seq):
        if is_protein(fp_fa_or_seq, alphabet):
            wt_seq = fp_fa_or_seq
            name_seq = args.output_id

            abnativ_vhh_humanisation(wt_seq, name_seq, args.nat_vh, args.nat_vhh, args.output_directory, args.pdb_file, args.ch_id,
                            fr_aho_indices, args.is_Exhaustive, args.threshold_abnativ_score, args.threshold_rasa_score,
                            args.perc_allowed_decrease_vhh, args.forbidden_mut, args.a, args.b, verbose=True)
        else: 
            raise Exception(f'Can not find the input file {fp_fa_or_seq} or if you gave a single protein sequence make sure it is composed only of {alphabet}')
    
    # If input is a fasta file
    else:
        for record in SeqIO.parse(fp_fa_or_seq, 'fasta'): 
            wt_seq = str(record.seq)
            name_seq = record.id

            abnativ_vhh_humanisation(wt_seq, name_seq, args.nat_vh, args.nat_vhh, args.output_directory, args.pdb_file, args.ch_id,
                            fr_aho_indices, args.is_Exhaustive, args.threshold_abnativ_score, args.threshold_rasa_score,
                            args.perc_allowed_decrease_vhh, args.forbidden_mut, args.a, args.b, verbose=True)

    
    

    








# (c) 2023 Sormannilab and Aubin Ramon
#
# Humanisation of VH/VL sequences using AbNatiV VH and VL assesments.
#
# ============================================================================

from .model.utils import is_protein
from .model.onehotencoder import alphabet
from .model.alignment.mybio import anarci_alignments_of_Fv_sequences_iter
from .model.scoring_functions import fr_aho_indices
from .humanisation.vh_vl_humanisation_functions import abnativ_vh_vl_humanisation
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq

import argparse

def run(args: argparse.Namespace):

    if is_protein(args.input_seq_vh, alphabet) and is_protein(args.input_seq_vl, alphabet):

        if '2' not in args.nat_vl:
            # Find Fv chain type to use the adequate unpaired model from AbNatiV v1
            seq_records = [SeqRecord(Seq(args.input_seq_vl), id='light_chain')]
            VH, VK, VL, failed, mischtype = anarci_alignments_of_Fv_sequences_iter(seq_records, isVHH=False, verbose=False)

            if len(VK)>0:
                nat_vl = 'VKappa'
            elif len(VL)>0:
                nat_vl = 'VLambda'
            else:
                raise ValueError(f'Could not compute automtically the light_Fv_type (VKappa, VLambda) of the input light sequence')

        else:
            nat_vl = args.nat_vl

        abnativ_vh_vl_humanisation(args.input_seq_vh, args.input_seq_vl, args.output_id, args.nat_vh, nat_vl, args.output_directory, args.pdb_file, args.ch_id_vh,
                                args.ch_id_vl, fr_aho_indices, fr_aho_indices, args.is_Exhaustive, args.threshold_abnativ_score, args.threshold_rasa_score, 
                                args.forbidden_mut, verbose=True)

    else: 
        raise Exception(f'The protein sequences you gave are not composed only of {alphabet}.')